from typing import Optional
from typing import Tuple

import attr
import numpy as np
from loguru import logger
from nptyping import assert_isinstance
from scipy import stats
from skimage import morphology

from avalon.common.errors import SwitchError
from avalon.datagen.errors import ImpossibleWorldError
from avalon.datagen.godot_base_types import FloatRange
from avalon.datagen.world_creation.configs.biome import generate_biome_config
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.world import WorldConfig
from avalon.datagen.world_creation.constants import HALF_AGENT_HEIGHT_VECTOR
from avalon.datagen.world_creation.debug_plots import plot_value_grid
from avalon.datagen.world_creation.entities.constants import CANONICAL_FOOD_HEIGHT_ON_TREE
from avalon.datagen.world_creation.entities.constants import FOOD_TREE_VISIBLE_HEIGHT
from avalon.datagen.world_creation.types import WorldType
from avalon.datagen.world_creation.worlds.difficulty import difficulty_variation
from avalon.datagen.world_creation.worlds.height_map import Point2DNP
from avalon.datagen.world_creation.worlds.height_map import Point3DNP
from avalon.datagen.world_creation.worlds.types import CompositionalConstraint
from avalon.datagen.world_creation.worlds.world import World
from avalon.datagen.world_creation.worlds.world_locations import WorldLocations


def _calculate_fractal_iteration_count(
    initial_count: int, grid_points_per_meter: float, min_size: float
) -> Tuple[int, float]:
    """:returns: how many times we need to double initial_count in order to exceed desired_count"""
    iterations = 0
    count = initial_count
    meters_per_grid_point = 1.0 / grid_points_per_meter
    while True:
        if count * meters_per_grid_point > min_size:
            break
        count *= 2
        iterations += 1
    return iterations, count * meters_per_grid_point


def _range_via_difficulty(min_val: float, max_val: float, difficulty: float, rand: np.random.Generator) -> float:
    delta = max_val - min_val
    return min_val + delta * rand.uniform() * difficulty


def _world_2d_to_3d(point_2d: Point2DNP, world: World) -> Point3DNP:
    assert_isinstance(point_2d, Point2DNP)
    index = world.map.point_to_index(point_2d)
    point_height = world.map.Z[index]
    return np.array([point_2d[0], point_height, point_2d[1]])


def create_world_from_constraint(
    goal_distance: stats.norm,
    rand: np.random.Generator,
    difficulty: float,
    export_config: ExportConfig,
    constraint: Optional[CompositionalConstraint],
    food_height: Optional[float] = None,
    min_size_in_meters: Optional[float] = None,
    ideal_shore_dist_options: Tuple[float, ...] = (4.0, 2.0),
) -> Tuple[World, WorldLocations]:
    if food_height is None:
        visibility_height = FOOD_TREE_VISIBLE_HEIGHT
        food_height = CANONICAL_FOOD_HEIGHT_ON_TREE
    else:
        visibility_height = food_height
    if constraint is None:
        world, locations = create_world_for_skill_scenario(
            rand,
            difficulty,
            food_height,
            goal_distance,
            export_config,
            min_size_in_meters=min_size_in_meters,
            ideal_shore_dist_options=ideal_shore_dist_options,
            visibility_height=visibility_height,
        )
        locations, world = world.begin_height_obstacles(locations)
    else:
        world, locations = constraint.world, constraint.locations
    return world, locations


# these are the default max and min sizes.. move diverse worlds are made larger
BASE_MIN_SIZE_IN_METERS = 30.0
BASE_MAX_SIZE_IN_METERS = 300.0
# how finely the height grid is specified. This corresponds to 0.5m per cell in the HeightMap
GRID_POINTS_PER_METER = 2.0
# how much space to flatten around spawn and goal locations.
# more diverse worlds will have a smaller flattening radius
MAX_FLATTEN_RADIUS = 5.0
MIN_FLATTEN_RADIUS = 1.0
# if visibility is required, how many goal locations to try before giving up and trying a different spawn location
# note that visibility is almost always required, except for some of the more complex and compositional tasks
MAX_VISIBILITY_CHECKS = 5
# how many different spawn locations to try before giving up and making an entirely new world.
# some worlds are simply broken given the constraints specified (ex: imagine asking for a goal and spawn point that
# are 10 meters apart, but there is only a single tiny patch of land)
MAX_SPAWN_TRIES_PER_MAP = 5
# how many different maps to try making before giving up.
# if this is reached, an ImpossibleWorldError is raised
# note that this normally means the caller is asking for constraints that are far too tight
MAX_OVERALL_RETRIES = 10
# prevent things from spawning in your body. No goal point will be selected that is closer than this to the spawn point
MIN_GOAL_DISTANCE = 1.0


def create_world_for_skill_scenario(
    rand: np.random.Generator,
    diversity: float,
    food_height: float,
    # worlds will usually be 2x this size so that you have a good chance of getting the right distance even if you randomly choose a point
    goal_distance: stats.norm,
    export_config: ExportConfig,
    is_logging: bool = False,
    world_type: Optional[WorldType] = None,
    is_visibility_required: bool = True,
    explicit_height_ratio: Optional[float] = None,
    foliage_density_modifier: Optional[float] = None,
    min_size_in_meters: Optional[float] = None,
    max_size_in_meters: Optional[float] = None,
    # first try spawning with us not right next to the water, then try without that
    ideal_shore_dist_options: Tuple[float, ...] = (4.0, 2.0),
    visibility_height: Optional[float] = None,
) -> Tuple[World, WorldLocations]:
    if visibility_height is None:
        visibility_height = food_height

    food_offset = np.array([0, food_height, 0])
    visibility_offset = np.array([0, visibility_height, 0])

    # TODO probably want to remove this
    is_debugging = False

    if max_size_in_meters is None:
        max_size_in_meters = BASE_MAX_SIZE_IN_METERS

    if foliage_density_modifier is None:
        foliage_density_modifier = difficulty_variation(5.0, 30.0, rand, diversity)

    # figure out how big the world should be
    goal_distance_preference = goal_distance.mean()
    # we won't return worlds where the goal is closer than two standard deviations from what you requested
    hard_min_distance = goal_distance_preference - goal_distance.std() * 2.0
    if hard_min_distance < 0.0:
        hard_min_distance = 0.0

    if is_debugging:
        logger.debug(f"Goal and spawn must be at least {hard_min_distance:.1f}m away")

    if min_size_in_meters is None:
        min_size_in_meters = goal_distance_preference * 2.0

    if min_size_in_meters < BASE_MIN_SIZE_IN_METERS:
        min_size_in_meters = BASE_MIN_SIZE_IN_METERS

    if is_debugging:
        logger.debug(f"Min size in meters: {min_size_in_meters}")
        logger.debug(f"Max size in meters: {max_size_in_meters}")

    if max_size_in_meters < min_size_in_meters:
        raise Exception(
            "This task is misconfigured. We should not be giving values that result in mins greater than maxes"
        )

    # we decrease diversity and retry if we fail to find points that are sufficiently far away and have visibility
    # this should generally take zero retries, so not really a performance concern
    num_visibility_check = 0
    for retry_num in range(MAX_OVERALL_RETRIES):

        # lower numbers are much higher diversity terrain. Min is 3 because we need to be able to interpolate
        initial_point_count = round(difficulty_variation(8, 3, rand, diversity))

        # how many points we want in the grid. Controls how many times we're going to iterate
        iteration_min_size_in_meters = min_size_in_meters
        fractal_iteration_count, size_in_meters = _calculate_fractal_iteration_count(
            initial_point_count, GRID_POINTS_PER_METER, iteration_min_size_in_meters
        )
        simplicity_warnings = []
        # if it's too big, just reduce the size and try again
        while size_in_meters > max_size_in_meters:
            iteration_min_size_in_meters /= 2.0
            fractal_iteration_count, size_in_meters = _calculate_fractal_iteration_count(
                initial_point_count, GRID_POINTS_PER_METER, iteration_min_size_in_meters
            )
        if size_in_meters < goal_distance_preference:
            simplicity_warnings.append(
                f"World was somewhat smaller than desired: {size_in_meters} vs {goal_distance_preference}"
            )

        if is_debugging:
            logger.info(f"World size: {size_in_meters}")
        # logger.info(f"World size: {size_in_meters}")

        # how much to flatten the world near the important points
        flatten_radius = (1.0 - diversity) * (MAX_FLATTEN_RADIUS - MIN_FLATTEN_RADIUS) + MIN_FLATTEN_RADIUS

        config = generate_world_config(
            rand,
            diversity,
            export_config,
            initial_point_count,
            fractal_iteration_count,
            size_in_meters,
            world_type,
            explicit_height_ratio,
        )
        biome_config = generate_biome_config(
            rand, export_config, config.final_max_altitude_meters, foliage_density_modifier
        )

        world = World.build(config, export_config, biome_config=biome_config, is_debug_graph_printing_enabled=False)
        map_new = world.map.copy()
        special_biomes_new = map_new.erode_shores(rand, biome_config)
        world = attr.evolve(world, special_biomes=special_biomes_new, map=map_new)

        if is_debugging and False:
            plot_value_grid(world.map.Z, "Possible world")
            world.map.plot()

        # don't spawn within two meters of the edge, that's asking for trouble
        edge_size = 2.0
        coord_range = FloatRange((-size_in_meters / 2.0) + edge_size, (size_in_meters / 2.0) - edge_size)
        for ideal_shore_dist in ideal_shore_dist_options:
            # pick a random point to spawn at
            for i in range(MAX_SPAWN_TRIES_PER_MAP):
                spawn_point_2d = world.map.get_random_land_point(rand, ideal_shore_dist)
                if not coord_range.contains(spawn_point_2d[0]) or not coord_range.contains(spawn_point_2d[1]):
                    if is_debugging:
                        logger.debug("Odd, out of range")
                    continue
                # the extra 1.1 is just to make sure you dont start in the ground :-P
                spawn_point = _world_2d_to_3d(spawn_point_2d, world) + HALF_AGENT_HEIGHT_VECTOR * 1.1

                # filter points to those on the same island
                island_mask, island_mask_without_fresh_water = world.map.get_island(spawn_point_2d)

                restricted_found = False
                if ideal_shore_dist > 0.0:
                    food_mask = island_mask
                    for multiple in (1.0, 0.5):
                        shore_cell_dist = int(multiple * world.map.cells_per_meter * ideal_shore_dist) + 1
                        # shrink the island a little bit so that we dont spawn on the shore because it's annoying
                        smaller_food_mask = np.logical_and(
                            island_mask,
                            np.logical_not(
                                morphology.dilation(np.logical_not(island_mask), morphology.disk(shore_cell_dist))
                            ),
                        )
                        if np.any(smaller_food_mask):
                            restricted_found = True
                            # plot_value_grid(island_mask)
                            # plot_value_grid(smaller_food_mask)
                            food_mask = smaller_food_mask
                            break
                else:
                    food_mask = island_mask

                # couldnt find points sufficiently far from the ocean, not good
                if not restricted_found:
                    if is_debugging:
                        logger.debug("not far enough from ocean")
                    continue

                # also prevent things from being just way too close
                if hard_min_distance > 0.0:
                    sq_min_dist = hard_min_distance**2
                    food_mask = np.logical_and(food_mask, world.map.get_dist_sq_to(spawn_point_2d) > sq_min_dist)

                possible_food_points = np.stack([world.map.X, world.map.Y], axis=2)[food_mask]

                # try a few randomly selected points (weighted by distance preference) until one is visible
                visibility_calculator = None
                if is_visibility_required:
                    visibility_calculator = world.map.generate_visibility_calculator()
                if goal_distance:
                    distances = np.linalg.norm(possible_food_points - spawn_point_2d, axis=1)
                    location_weights = goal_distance.pdf(distances)
                    location_weights[distances < MIN_GOAL_DISTANCE] = 0.0
                    total_weights = location_weights.sum()
                    if total_weights <= 0.0:
                        if is_debugging:
                            logger.debug("No likely food locations")
                        continue
                    location_probabilities = location_weights / total_weights
                    food_point_2ds = rand.choice(possible_food_points, MAX_VISIBILITY_CHECKS, p=location_probabilities)
                else:
                    food_point_2ds = rand.choice(possible_food_points, MAX_VISIBILITY_CHECKS)
                for food_point_2d in food_point_2ds:
                    if not coord_range.contains(food_point_2d[0]) or not coord_range.contains(food_point_2d[1]):
                        if is_debugging:
                            logger.debug("food point is out of range")
                        continue
                    # can't spawn food in literally the same spot as you :-P
                    if np.isclose(food_point_2d, spawn_point_2d).all():
                        if is_debugging:
                            logger.debug("Food is too close to spawn")
                        continue
                    if visibility_calculator is None:
                        # we disabled visibility above, so dont worry about it
                        is_food_visible = True
                    else:
                        # adds a bit of extra height because the spawn point is technically the middle of the agent
                        is_food_visible = visibility_calculator.is_visible_from(
                            spawn_point + HALF_AGENT_HEIGHT_VECTOR,
                            _world_2d_to_3d(food_point_2d, world) + visibility_offset,
                        )
                    if is_food_visible:
                        # how much space to reserve around each object
                        object_radius = 0.75
                        # if they're both within the flattening region, just flatten it all out
                        final_distance = np.linalg.norm(spawn_point_2d - food_point_2d)
                        if final_distance < object_radius * 2.0:
                            world = world.flatten(
                                (spawn_point_2d + food_point_2d) / 2.0, flatten_radius, object_radius * 3.0
                            )
                        else:
                            flatten_radius = min([flatten_radius, final_distance / 2.0])  # type: ignore
                            world = world.flatten(spawn_point_2d, flatten_radius, object_radius)
                            world = world.flatten(food_point_2d, flatten_radius, object_radius)
                        if is_logging:
                            logger.info(f"Visibility checks: {num_visibility_check}")
                        return world, WorldLocations(
                            island=island_mask_without_fresh_water,
                            spawn=spawn_point,
                            goal=_world_2d_to_3d(food_point_2d, world) + food_offset,
                        )
                    else:
                        if is_debugging:
                            logger.debug("food is not visible")
                        num_visibility_check += 1

    raise ImpossibleWorldError("Something is very wrong, or we got very unlucky...")


# worlds smaller than this will be boosted out of the water by the below functions.
# helps ensure there is a sufficiently large land mass on which we can spawn things.
MIN_WORLD_SIZE_FOR_COMPLEXITY = 40.0


def generate_world_config(
    rand: np.random.Generator,
    diversity: float,
    export_config: Optional[ExportConfig],
    # terrain construction works by iteratively subdividing a grid.
    # this controls the initial size of the grid.
    # lower numbers are much higher diversity terrain because the iteration works by adding noise, then subdividing.
    # Min is 3 because we need to be able to interpolate
    initial_point_count: int,
    # how many times to do the add noise + subdivide loop
    fractal_iteration_count: int,
    size_in_meters: float,
    world_type: Optional[WorldType],
    explicit_height_ratio: Optional[float],
) -> WorldConfig:
    # how fine grained the triangulation is
    # 4.0 is ultra high quality. Almost looks like a texture at that point
    # 1.0 is very high quality.
    # 0.1 is medium quality. You can feel that the ground is triangulated, but it doesn't feel offensive
    # 0.05 is about as low as you can go without things feeling broken
    point_density_in_points_per_square_meter = 0.1
    # helps wiggle the edge of the world
    # at 100m, 0.005 is pleasing when looking at the heightmaps
    fade_noise_scale = 0.005 * (100.0 / size_in_meters)
    # there is always a chance, for these simple worlds, that we create a fairly simple world
    # this allows us to practice the skills in a sort of ideal, platonic setting
    platonic_world_probability = np.interp(diversity, [0.0, 0.2, 0.6, 1.0], [1.0, 0.9, 0.4, 0.2])
    is_platonic_world = bool(rand.random() < platonic_world_probability)
    if world_type is not None:
        if world_type == WorldType.PLATONIC:
            is_platonic_world = True
        else:
            is_platonic_world = False
    # what really defines a platonic world is the noise scale of the mountains relative to the other noise
    # more mountains also makes it more platonic... (if they're in the same spot)
    if is_platonic_world:
        is_mountain_placement_normal_distribution = False
        mountain_noise_count = 1
        mountain_radius_decay = 1.0
        noise_scale_decay = pow(10.0, rand.uniform(-0.5, 0.5))
        mountain_radius = _range_via_difficulty(0.6, 0.4, diversity, rand)
        mountain_noise_scale = _range_via_difficulty(5.0, 3.0, diversity, rand)
        mountain_offset = _range_via_difficulty(0.45, 0.3, diversity, rand)
        smoothing_probability = np.interp(diversity, [0.0, 0.2, 0.6, 1.0], [0.5, 0.4, 0.2, 0.1])
        blur_meters_std_dev = 0.0
        is_smoothed = bool(rand.random() < smoothing_probability)
        smooth_min = 0.0
        iterated_noise_scale = pow(noise_scale_decay, fractal_iteration_count)
        if iterated_noise_scale > 15:
            is_smoothed = True
            smooth_min = 1.0
        if is_smoothed:
            blur_meters_std_dev = _range_via_difficulty(1.5, smooth_min, diversity, rand)
        height_ratio = _range_via_difficulty(0.05, 0.2, diversity, rand)
    else:
        if world_type is None:
            # TODO: maybe put archipelagos and stuff back someday, they interfere with the generation too much for now
            # # randomly select the type of world
            # world_type = rand.choice(
            #     [WorldType.CONTINENT, WorldType.ARCHIPELAGO, WorldType.JAGGED],
            #     p=[0.8, 0.2, 0.0],
            # )
            # logger.info(world_type)
            world_type = WorldType.CONTINENT

        # jagged noisy worlds
        if world_type == WorldType.JAGGED:
            # A value of 0.01 here is particularly pleasing
            # This is the scale at which our last iteration ends up
            iterated_noise_scale = pow(0.2, rand.normal(2.0, 0.5))
            noise_scale_decay = iterated_noise_scale ** (1 / (fractal_iteration_count - 1))
            is_mountain_placement_normal_distribution = False
            mountain_noise_count = 0
            mountain_radius = 0.0
            mountain_noise_scale = 0.0
            mountain_offset = 0.0
            mountain_radius_decay = 0.0
            # when small, boost it out of the water
            if size_in_meters < MIN_WORLD_SIZE_FOR_COMPLEXITY:
                mountain_noise_count = 1
                mountain_radius = 0.5
                mountain_noise_scale = 1.0
                mountain_offset = 0.4
                mountain_radius_decay = 1.0
            blur_meters_std_dev = 0.0
            if iterated_noise_scale > 0.1:
                blur_meters_std_dev = _range_via_difficulty(1.0, 0.0, diversity, rand)
            min_height_ratio = 0.02
            height_ratio_mean = (min_height_ratio + diversity * 0.1) + 0.05
            height_ratio_std_dev = (height_ratio_mean - min_height_ratio) * 0.5 * diversity
            height_ratio = rand.normal(height_ratio_mean, height_ratio_std_dev)
            height_ratio = max(height_ratio, min_height_ratio)
        elif world_type == WorldType.ARCHIPELAGO:
            height_variation = rand.normal()
            height_mean = 0.05 + diversity * 0.05
            if height_variation < 0:
                height_variation *= height_mean / 3.0
            else:
                height_variation *= 0.07 * diversity
            height_ratio = height_mean + height_variation
            min_height_ratio = 0.02
            height_ratio = max(height_ratio, min_height_ratio)

            noise_variation = rand.normal()
            if noise_variation < 0:
                noise_variation *= 2
            else:
                noise_variation *= 7 * diversity
            # 20 meters of noise is pretty difficult to traverse. You spend much of your time climbing
            #  roughly corresponds to very rocky outcroppings
            # 2 meters of noise end up with fairly flat terrain in large worlds
            meters_of_noise = max(6.0 + noise_variation, 0.5)
            noise_scale_base = meters_of_noise / (size_in_meters * height_ratio)

            # A value of 0.01 here is particularly pleasing
            # This is the scale at which our last iteration ends up
            iterated_noise_scale = pow(noise_scale_base, rand.normal(2.0, 0.5))
            noise_scale_decay = iterated_noise_scale ** (1 / (fractal_iteration_count - 1))
            is_mountain_placement_normal_distribution = False
            mountain_noise_count = 0
            mountain_radius = 0.0
            mountain_noise_scale = 0.0
            mountain_offset = 0.0
            mountain_radius_decay = 0.0
            # when small, boost it out of the water
            if size_in_meters < MIN_WORLD_SIZE_FOR_COMPLEXITY:
                mountain_noise_count = 1
                mountain_radius = 0.5
                mountain_noise_scale = 1.0
                mountain_offset = 0.4
                mountain_radius_decay = 1.0
            blur_meters_std_dev = 0.0
            if iterated_noise_scale > 0.04:
                blur_meters_std_dev = _range_via_difficulty(2.0, 0.2, diversity, rand)
        elif world_type == WorldType.CONTINENT:
            height_variation = rand.normal()
            height_mean = 0.05 + diversity * 0.1
            if height_variation < 0:
                height_variation *= height_mean / 3.0
            else:
                height_variation *= 0.07 * diversity
            height_ratio = height_mean + height_variation
            min_height_ratio = 0.02
            height_ratio = max(height_ratio, min_height_ratio)

            noise_variation = rand.normal()
            if noise_variation < 0:
                noise_variation *= 2
            else:
                noise_variation *= 7 * diversity
            # 20 meters of noise is pretty difficult to traverse. You spend much of your time climbing
            #  roughly corresponds to very rocky outcroppings
            # 2 meters of noise end up with fairly flat terrain in large worlds
            meters_of_noise = max(6.0 + noise_variation, 0.5)
            noise_scale_base = meters_of_noise / (size_in_meters * height_ratio)

            # A value of 0.01 here is particularly pleasing
            # This is the scale at which our last iteration ends up
            iterated_noise_scale = pow(noise_scale_base, rand.normal(2.0, 0.5))
            noise_scale_decay = iterated_noise_scale ** (1 / (fractal_iteration_count - 1))

            mountain_mode = rand.integers(0, 2)
            if mountain_mode == 0:
                # make one large central mountain
                is_mountain_placement_normal_distribution = rand.uniform() < 0.5
                mountain_noise_count = rand.integers(8, 12)
                mountain_radius_decay = rand.uniform(0.2, 0.7)
                mountain_radius = 0.4
                mountain_noise_scale = 0.1
                mountain_offset = _range_via_difficulty(0.45, 0.35, diversity, rand)
            elif mountain_mode == 1:
                # continually raise the middle section of the map out of the water
                # to make one relatively large continent
                is_mountain_placement_normal_distribution = False
                mountain_noise_count = 2
                mountain_radius_decay = 1.0
                mountain_radius = _range_via_difficulty(0.6, 0.2, diversity, rand)
                # controls how much it ends up being a single landmass
                # if the noise scale is higher, we end up with more of a single mass
                mountain_noise_scale = 0.12 + rand.normal() * 0.03
                mountain_offset = _range_via_difficulty(0.45, 0.2, diversity, rand)
            else:
                raise SwitchError(f"Unhandled mountain mode: {mountain_mode}")

            # when small, boost it out of the water
            if size_in_meters < MIN_WORLD_SIZE_FOR_COMPLEXITY:
                is_mountain_placement_normal_distribution = False
                mountain_noise_count = 1
                mountain_radius = 0.5
                mountain_noise_scale = 1.0
                mountain_offset = 0.4
                mountain_radius_decay = 1.0
            blur_meters_std_dev = 0.0
            if iterated_noise_scale > 0.04:
                blur_meters_std_dev = _range_via_difficulty(2.0, 0.2, diversity, rand)
        else:
            raise SwitchError(f"Unhandled world type: {world_type}")

    # just for debugging:
    if explicit_height_ratio is not None:
        height_ratio = explicit_height_ratio
    final_max_altitude_meters = height_ratio * size_in_meters
    # see the definition of this object for a description of each parameter
    config = WorldConfig(
        seed=rand.integers(0, np.iinfo(np.int64).max),
        point_density_in_points_per_square_meter=point_density_in_points_per_square_meter,
        size_in_meters=size_in_meters,
        initial_point_count=initial_point_count,
        fractal_iteration_count=fractal_iteration_count,
        # the scale of the noise actually doesn't matter because we reset the height afterwards
        # only really matters relative to mountain_noise_scale
        initial_noise_scale=1.0,
        noise_scale_decay=noise_scale_decay,
        is_mountain_placement_normal_distribution=is_mountain_placement_normal_distribution,
        mountain_noise_count=mountain_noise_count,
        mountain_radius=mountain_radius,
        mountain_noise_scale=mountain_noise_scale,
        mountain_offset=mountain_offset,
        mountain_radius_decay=mountain_radius_decay,
        blur_meters_std_dev=blur_meters_std_dev,
        final_max_altitude_meters=final_max_altitude_meters,
        final_world_edge_fade_fraction=0.01,
        fade_fraction=0.2,
        fade_noise_scale=fade_noise_scale,
        fade_noise_min=0.1,
        # made most worlds circular because it's prettier :-P
        is_fade_circular=rand.uniform() < 0.75,
    )
    return config



import warnings
from typing import Any
from typing import List
from typing import Tuple
from typing import cast

import gym
import numpy as np
from gym.wrappers import TimeLimit
from loguru import logger

from avalon.agent.common import wrappers
from avalon.agent.common.params import DmcEnvironmentParams
from avalon.agent.common.params import EnvironmentParams
from avalon.agent.common.params import ProcgenEnvironmentParams
from avalon.agent.common.test_envs import TestEnvironmentParams
from avalon.agent.common.wrappers import ElapsedTimeWrapper
from avalon.common.type_utils import assert_not_none


def build_env(env_params: EnvironmentParams, torchify: bool = True) -> gym.Env:
    # TODO: I should add a "checker" wrapper that checks that observations and actions match that specced in the Space.
    seed = env_params.env_index
    env: gym.Env
    if env_params.suite == "dmc":
        # For rendering, you'll need libglew-(dev/2.0/2.1), and MUJOCO_GL=egl
        assert isinstance(env_params, DmcEnvironmentParams)
        assert env_params.task is not None
        assert "-" not in env_params.task, "use underscores in task names"
        # This camera config comes from the dreamerv2 repo
        camera = dict(
            quadruped_walk=2,
            quadruped_run=2,
            quadruped_escape=2,
            quadruped_fetch=2,
            locom_rodent_maze_forage=1,
            locom_rodent_two_touch=1,
        ).get(env_params.task, 0)
        env = DeepMindControl(
            env_params.task,
            size=(64, 64),
            include_state=env_params.include_proprio,
            include_rgb=env_params.include_rgb,
            camera=camera,
            seed=seed,
        )
        env = wrappers.RecordEpisodeStatistics(env)  # must happen before reward scaling
        env = wrappers.ActionRepeat(env, env_params.action_repeat)
        # rescales actions from standard ranges to the envs desired range.
        if env_params.time_limit:
            env = wrappers.TimeLimit(env, max_episode_steps=env_params.time_limit // env_params.action_repeat)
        env = wrappers.DictObsActionWrapper(env)
        if env_params.include_rgb:
            env = wrappers.ImageTransformWrapper(env, key="rgb", greyscale=False, resolution=None)
    elif env_params.suite == "godot":
        from avalon.agent.godot.godot_gym import AvalonEnv
        from avalon.agent.godot.godot_gym import GodotEnvironmentParams
        from avalon.agent.godot.godot_gym import GodotObsTransformWrapper
        from avalon.agent.godot.godot_gym import ScaleAndSquashAction

        assert isinstance(env_params, GodotEnvironmentParams)
        assert env_params.time_limit is None, "godot has its own time limit"
        assert env_params.action_repeat == 1

        # Note: This will seed itself properly using env_index
        env = AvalonEnv(env_params)
        # We don't use the TimeLimit wrapper because the time limit is dynamic,
        # so we trust that the godot env gives the proper TimeLimit.truncated signal
        # (which it should) for the timelimit boostrapping to work properly if enabled.
        env = GodotObsTransformWrapper(env, greyscale=env_params.greyscale)
        if env_params.mode == "train":
            from avalon.agent.godot.godot_gym import CurriculumWrapper

            env = CurriculumWrapper(
                env,  # type: ignore[arg-type]
                task_difficulty_update=env_params.task_difficulty_update,
                meta_difficulty_update=env_params.meta_difficulty_update,
            )
        env = ScaleAndSquashAction(env, scale=1)
        env = wrappers.OneHotActionWrapper(env)
        # env = RewardSoftClipWrapper(env, scale=5)
    elif env_params.suite == "test":
        # Note: haven't implemented proper seeding in these test envs.
        assert env_params.action_repeat == 1
        from avalon.agent.common.test_envs import get_env

        assert type(env_params) == TestEnvironmentParams
        env = get_env(env_params.task, env_params)
        env = wrappers.DictObsActionWrapper(env)
        env = wrappers.OneHotActionWrapper(env)
        env = wrappers.RecordEpisodeStatistics(env)  # must happen before reward scaling
        if env_params.time_limit:
            env = wrappers.TimeLimit(env, max_episode_steps=env_params.time_limit // env_params.action_repeat)
    elif env_params.suite == "gym":
        assert env_params.action_repeat == 1
        # Annoyingly, gym envs apply their own time limit already.
        logger.info("time limit arg ignored in gym envs")
        env = gym.make(assert_not_none(env_params.task))
        env.seed(seed)
        # Hacky. Relies on the TimeWrapper being the outermost wrapper. Not sure the better way.
        assert isinstance(env, (ElapsedTimeWrapper, TimeLimit))
        max_steps = env._max_episode_steps
        logger.info(f"env has a time limit of {max_steps} steps")
        if env_params.pixel_obs_wrapper:
            env = wrappers.PixelObsWrapper(env)
            env = wrappers.DictObsActionWrapper(env, obs_key="rgb")
        else:
            env = wrappers.DictObsActionWrapper(env, obs_key="state")  # type: ignore
        if env_params.pixel_obs_wrapper:
            env = wrappers.ImageTransformWrapper(env, key="rgb", greyscale=True, resolution=64)
        env = wrappers.OneHotActionWrapper(env)
        if env_params.elapsed_time_obs:
            env = wrappers.ElapsedTimeWrapper(env, max_steps)
        env = wrappers.RecordEpisodeStatistics(env)  # must happen before reward scaling
    elif env_params.suite == "atari":
        assert env_params.task is not None
        assert env_params.action_repeat == 4
        assert env_params.time_limit == 27000
        assert env_params.elapsed_time_obs is False
        # These are the settings from dreamerv2
        env = Atari(env_params.task, action_repeat=env_params.action_repeat, size=(64, 64), grayscale=True)
        env.seed(seed)
        if env_params.time_limit:
            # danijar applies the time limit in agent-steps, not action-repeated env steps
            env = wrappers.TimeLimit(env, max_episode_steps=env_params.time_limit)
        env = wrappers.DictObsActionWrapper(env, obs_key="rgb")
        env = wrappers.OneHotActionWrapper(env)
        env = wrappers.RecordEpisodeStatistics(env)  # must happen before reward scaling
        # Note the tanh here!
        env = wrappers.ScaleRewards(env, func=np.tanh)
        # Just converts from hwc to chw
        env = wrappers.ImageTransformWrapper(env, key="rgb")
    elif env_params.suite == "procgen":
        warnings.filterwarnings("ignore", message=".*Future gym versions will require.*")
        # Need this import to register the procgen envs?
        import procgen  # isort: skip

        assert isinstance(env_params, ProcgenEnvironmentParams)
        assert env_params.task is not None
        assert env_params.time_limit is None, "procgen has its own time limits (altho they don't set the info flags)"

        env = gym.make(
            f"procgen-{env_params.task}-v0",
            start_level=env_params.start_level,
            num_levels=env_params.num_levels,
            distribution_mode=env_params.distribution_mode,
            rand_seed=env_params.env_index,
        )
        env = wrappers.DictObsActionWrapper(env, obs_key="rgb")
        env = wrappers.OneHotActionWrapper(env)
        env = wrappers.RecordEpisodeStatistics(env)  # must happen before reward scaling
        env = gym.wrappers.NormalizeReward(env)
        env = gym.wrappers.TransformReward(env, lambda reward: np.clip(reward, -10, 10))  # type: ignore[no-any-return]
        # Just converts from hwc to chw
        env = wrappers.ImageTransformWrapper(env, key="rgb")
    else:
        assert False
    env = wrappers.NormalizeActions(env)
    env = wrappers.ScaleRewards(env, env_params.reward_scale)
    env = wrappers.ClipActionWrapper(env)
    if env_params.frame_stack != 1:
        env = wrappers.DictFrameStack(env, num_stack=env_params.frame_stack)
    if torchify:
        # My worker requires this, but it won't work with eg the builtin gym vecenv.
        env = wrappers.Torchify(env)
    return env


DMC_TASKS = [
    "acrobot_swingup",
    "cartpole_balance",
    "cartpole_balance_sparse",
    "cartpole_swingup",
    "cartpole_swingup_sparse",
    "cheetah_run",
    "cup_catch",
    "finger_spin",
    "finger_turn_easy",
    "finger_turn_hard",
    "hopper_hop",
    "hopper_stand",
    "pendulum_swingup",
    "quadruped_run",
    "quadruped_walk",
    "reacher_easy",
    "reacher_hard",
    "walker_walk",
    "walker_stand",
    "walker_run",
]

PROCGEN_ENVS = [
    "coinrun",
    "starpilot",
    "caveflyer",
    "dodgeball",
    "fruitbot",
    "chaser",
    "miner",
    "jumper",
    "leaper",
    "maze",
    "bigfish",
    "heist",
    "climber",
    "plunder",
    "ninja",
    "bossfight",
]


class DeepMindControl(gym.Env):
    def __init__(
        self,
        name: str,
        size: Tuple[int, int] = (64, 64),
        camera: Any = None,
        include_state: bool = False,
        include_rgb: bool = True,
        seed: int = 0,
    ) -> None:
        from dm_control import suite

        domain, task = name.split("_", 1)
        if domain == "cup":  # Only domain with multiple words.
            domain = "ball_in_cup"
        self._env = suite.load(domain, task, task_kwargs={"random": seed})
        self._size = size
        if camera is None:
            camera = dict(quadruped=2).get(domain, 0)
        self._camera = camera
        self.include_state = include_state
        self.include_rgb = include_rgb
        # TODO: fix!
        # We just ignore all scalar spaces, because this is how danijar did it (presumably accidentally).
        self.scalar_spaces: List[str] = []

    @property
    def observation_space(self) -> gym.spaces.Dict:  # type: ignore[override]
        spaces: dict[str, gym.Space] = {}
        if self.include_state:
            for key, value in self._env.observation_spec().items():
                logger.warning("gym spaces do not give observation ranges. no rescaling will be applied.")
                if value.shape == ():
                    self.scalar_spaces.append(key)
                    continue
                spaces[key] = gym.spaces.Box(-np.inf, np.inf, value.shape, dtype=np.float32)
        if self.include_rgb:
            spaces["rgb"] = gym.spaces.Box(0, 255, self._size + (3,), dtype=np.uint8)
        return gym.spaces.Dict(spaces)

    @property
    def action_space(self) -> gym.spaces.Box:  # type: ignore[override]
        spec = self._env.action_spec()
        return gym.spaces.Box(spec.minimum, spec.maximum, dtype=np.float32)

    def step(self, action):  # type: ignore
        time_step = self._env.step(action)
        obs = {}
        if self.include_state:
            state = {k: v for k, v in dict(time_step.observation).items() if k not in self.scalar_spaces}
            state = {k: v.astype(np.float32) if v.dtype == np.float64 else v for k, v in state.items()}
            obs |= state
        if self.include_rgb:
            obs["rgb"] = self.render()
        reward = time_step.reward or 0
        done = time_step.last()
        info = {"discount": np.array(time_step.discount, np.float32)}
        return obs, reward, done, info

    def reset(self):  # type: ignore
        time_step = self._env.reset()
        obs = {}
        if self.include_state:
            state = {k: v for k, v in dict(time_step.observation).items() if k not in self.scalar_spaces}
            state = {k: v.astype(np.float32) if v.dtype == np.float64 else v for k, v in state.items()}
            obs |= state
        if self.include_rgb:
            obs["rgb"] = self.render()
        return obs

    def render(self, *args, **kwargs):  # type: ignore
        if kwargs.get("mode", "rgb_array") != "rgb_array":
            raise ValueError("Only render mode 'rgb_array' is supported.")
        return self._env.physics.render(*self._size, camera_id=self._camera)


ATARI_TASKS = [
    # "adventure",
    # "air_raid",
    "alien",
    "amidar",
    "assault",
    "asterix",
    "asteroids",
    "atlantis",
    "bank_heist",
    "battle_zone",
    "beam_rider",
    "berzerk",
    "bowling",
    "boxing",
    "breakout",
    # "carnival",
    "centipede",
    "chopper_command",
    "crazy_climber",
    # "defender",
    "demon_attack",
    "double_dunk",
    # "elevator_action",
    "enduro",
    "fishing_derby",
    "freeway",
    "frostbite",
    "gopher",
    "gravitar",
    "hero",
    "ice_hockey",
    "jamesbond",
    # "journey_escape",
    # "kaboom",
    "kangaroo",
    "krull",
    "kung_fu_master",
    "montezuma_revenge",
    "ms_pacman",
    "name_this_game",
    "phoenix",
    "pitfall",
    "pong",
    # "pooyan",
    "private_eye",
    "qbert",
    "riverraid",
    "road_runner",
    "robotank",
    "seaquest",
    "skiing",
    "solaris",
    "space_invaders",
    "star_gunner",
    "tennis",
    "time_pilot",
    "tutankham",
    "up_n_down",
    "venture",
    "video_pinball",
    "wizard_of_wor",
    "yars_revenge",
    "zaxxon",
]


class Atari(gym.Env):
    def __init__(
        self,
        name: str,
        action_repeat: int = 4,
        size: Tuple[int, int] = (84, 84),
        grayscale: bool = True,
        noops: int = 30,
        life_done: bool = False,
        sticky: bool = True,
        all_actions: bool = False,
    ) -> None:
        assert size[0] == size[1]
        import gym.envs.atari
        import gym.wrappers

        if name == "james_bond":
            name = "jamesbond"
        # this source is in https://github.com/mgbellemare/Arcade-Learning-Environment/blob/master/src/gym/envs/atari/environment.py
        env = gym.envs.atari.AtariEnv(
            game=name,
            obs_type="image",
            frameskip=1,
            repeat_action_probability=0.25 if sticky else 0.0,
            full_action_space=all_actions,
        )
        # Avoid unnecessary rendering in inner env.
        env._get_obs = lambda: None  # type: ignore
        # Tell wrapper that the inner env has no action repeat.
        env.spec = gym.envs.registration.EnvSpec("NoFrameskip-v0")
        self._env = gym.wrappers.AtariPreprocessing(env, noops, action_repeat, size[0], life_done, grayscale)
        self._size = size
        self._grayscale = grayscale

    @property
    def observation_space(self) -> gym.spaces.Box:  # type: ignore[override]
        shape = self._size + (1 if self._grayscale else 3,)
        return gym.spaces.Box(0, 255, shape, np.uint8)

    @property
    def action_space(self) -> gym.spaces.Box:  # type: ignore[override]
        return cast(gym.spaces.Box, self._env.action_space)

    def step(self, action: int):  # type: ignore
        image, reward, done, info = self._env.step(action)
        if self._grayscale:
            image = image[..., None]
        # info["is_terminal"] = done
        # info["is_first"] = False
        # info["is_last"] = done
        # image = rearrange(image, "h w c -> c h w")
        return image, reward, done, info

    def reset(self):  # type: ignore
        image = self._env.reset()
        if self._grayscale:
            image = image[..., None]
        # image = rearrange(image, "h w c -> c h w")
        return image

    def close(self):  # type: ignore
        return self._env.close()

    def render(self, mode: str = "human"):
        raise NotImplementedError



import time

import attr
from loguru import logger

from avalon.agent.common.parse_args import parse_args
from avalon.agent.common.trainer import OffPolicyTrainer
from avalon.agent.common.util import setup_new_process
from avalon.agent.dreamer.params import DreamerParams
from avalon.agent.godot.godot_eval import test
from avalon.agent.godot.godot_gym import GodotEnvironmentParams
from avalon.agent.godot.godot_gym import TrainingProtocolChoice
from avalon.common.error_utils import capture_exception
from avalon.common.log_utils import configure_remote_logger
from avalon.datagen.godot_env.interactive_godot_process import GODOT_ERROR_LOG_PATH

FRAGMENT_LENGTH = 30


@attr.s(auto_attribs=True, frozen=True)
class DreamerGodotParams(DreamerParams):
    total_env_steps: int = 50_000_000
    replay_buffer_size_timesteps: int = 1_000_000
    pcont: bool = False
    prioritize_ends: bool = False
    worker_managers: int = 1
    num_workers: int = 16
    discount: float = 0.99
    model_lr: float = 1e-4
    value_lr: float = 1e-4
    actor_lr: float = 1e-5
    clip_grad_norm: float = 100
    env_params: GodotEnvironmentParams = GodotEnvironmentParams(
        task_difficulty_update=6e-4,
        energy_cost_coefficient=1e-8,
        training_protocol=TrainingProtocolChoice.MULTI_TASK_BASIC,
        fixed_world_max_difficulty=0.5,
        # Make the "eating" always be part of the final fragment, so the end is at least theoretically predictable.
        # Will make episodes where the reward is gotten in < 2 frames be skipped, though.
        num_frames_alive_after_food_is_gone=FRAGMENT_LENGTH - 2,
        gpu_id=1,
        test_episodes_per_task=51,
    )
    batch_size: int = 116
    log_freq_hist: int = 2000
    log_freq_scalar: int = 25
    log_freq_media: int = 500
    checkpoint_every: int = 20000
    free_nats: float = 0
    kl_loss_scale: float = 30
    policy_entropy_scale: float = 2e-3
    pcont_loss_scale: float = 10
    reward_loss_scale: float = 10
    kl_balance: float = 0.8
    disclam: float = 0.95
    freeze_actor_steps: int = 500
    min_fragment_len: int = FRAGMENT_LENGTH
    max_fragment_len: int = FRAGMENT_LENGTH
    train_gpu: int = 0
    inference_gpus: tuple[int, ...] = (1,)
    actor_grad: str = "reinforce"
    log_rollout_metrics_every: int = 100
    is_training: bool = True
    is_testing: bool = False
    num_dataloader_workers: int = 1
    rssm_hidden_size: int = 600
    deter_size: int = 800
    stoch_size: int = 32
    center_and_clamp_discrete_logits = True


def run(params: DreamerParams):
    trainer = OffPolicyTrainer(params)
    try:
        if params.is_training:
            logger.info(f"starting training at {time.time()}")
            trainer.start_train_rollouts()
            trainer.train()
            trainer.shutdown_train_rollouts()
            logger.info(f"finished training at {time.time()}")
        trainer.shutdown()
        del trainer.train_dataloader
        if params.is_testing:
            logger.info(f"starting eval at {time.time()}")
            test(trainer.params, trainer.algorithm, log=True)
            logger.info(f"finished eval at {time.time()}")
    except Exception as e:
        if not params.debug:
            capture_exception(e)
        raise
    finally:
        logger.info("shutting down")
        if not params.debug:
            try:
                trainer.wandb_run.save(f"{GODOT_ERROR_LOG_PATH}/*")
            except Exception as e:
                logger.warning(f"wandb godot tarball save failed: {e}")
        trainer.shutdown(finish_wandb_quietly=True)


if __name__ == "__main__":
    configure_remote_logger(level="INFO")
    setup_new_process()
    try:
        default_params = DreamerGodotParams()
        default_params = parse_args(default_params)
        run(default_params)
    except Exception as e:
        logger.exception(e)
        raise



from abc import ABC
from abc import abstractmethod

# ObsType = Dict[str, Tensor]
from typing import Any
from typing import Dict
from typing import Generic
from typing import List
from typing import Tuple
from typing import TypeVar

import attr
import gym
import torch
from torch import Tensor

from avalon.agent.common.params import Params
from avalon.agent.common.util import pack_1d_list
from avalon.agent.common.util import pack_2d_list
from avalon.agent.common.util import seed_and_run_deterministically_if_enabled

Info = Dict[str, Any]
Observation = Dict[str, Tensor]  # represents a single timestep. atoms can have any shape
ObservationBatch = Dict[str, Tensor]  # shape (batch_size, ...)
ObservationSequenceBatch = Dict[str, Tensor]  # shape (batch_size, num_timesteps, ...)
ObservationSequence = Dict[str, Tensor]  # shape (num_timesteps, ...)

# ActionType = Dict[str, Tensor]
Action = Dict[str, Tensor]  # represents a single timestep. atoms can have any shape
ActionBatch = Dict[str, Tensor]  # shape (batch_size, ...)
ActionSequenceBatch = Dict[str, Tensor]  # shape (batch_size, num_timesteps, ...)
ActionSequence = Dict[str, Tensor]  # shape (num_timesteps, ...)

# The state is like {"stoch": tensor, "deter": tensor}
LatentBatch = Dict[str, Tensor]
StateActionBatch = Tuple[LatentBatch, ActionBatch]


@attr.s(auto_attribs=True, frozen=True)
class SequenceData:
    # All atoms should have shape (timesteps, ...)
    observation: ObservationSequence
    action: ActionSequence
    reward: Tensor
    done: Tensor
    is_terminal: Tensor


@attr.s(auto_attribs=True, frozen=True)
class BatchSequenceData:
    # All atoms should have shape (batch_size, num_timesteps, ...)
    observation: ObservationSequenceBatch
    action: ActionSequenceBatch
    reward: Tensor
    done: Tensor
    is_terminal: Tensor


@attr.s(auto_attribs=True, frozen=True)
class BatchData:
    # All atoms should have shape (batch_size, ...)
    observation: ObservationBatch
    action: ActionBatch
    reward: Tensor
    done: Tensor
    is_terminal: Tensor


@attr.s(auto_attribs=True, frozen=True)
class StepData:
    # if obs_first=True, for a given timestep, the order is (observation, action, reward/done/info)
    # if obs_first=False, for a given timestep, the order is (action, reward/done/info, observation)
    # if we extend this with extra inference fields, they'll go with the obs, since it's computed from the obs.
    # So eg for PPO, obs_first=True has ordering (observation/value/policy, action, reward/done/info)
    batch_type = BatchData
    sequence_type = SequenceData
    batch_sequence_type = BatchSequenceData

    observation: Dict[str, Tensor]
    action: Dict[str, Tensor]
    reward: float
    done: bool
    is_terminal: bool
    info: Dict[str, Any]

    def pack_sequence_batch(self, batch: List[List["StepData"]]) -> "BatchSequenceData":
        return pack_2d_list(batch, out_cls=self.batch_sequence_type)  # type: ignore

    def pack_sequence(self, sequence: List["StepData"]) -> "SequenceData":
        return pack_1d_list(sequence, out_cls=self.sequence_type)  # type: ignore


ParamsType = TypeVar("ParamsType", bound=Params)


class Algorithm(torch.nn.Module, ABC, Generic[ParamsType]):
    step_data_type: StepData = StepData

    def __init__(self, params: ParamsType, obs_space: gym.spaces.Dict, action_space: gym.spaces.Dict) -> None:
        seed_and_run_deterministically_if_enabled()

        super().__init__()
        self.params = params
        self.obs_space = obs_space
        self.action_space = action_space

    @abstractmethod
    def rollout_step(
        self,
        next_obs: Observation,
        dones: Tensor,
        indices_to_run: Tensor,
        exploration_mode: str,
    ) -> Tuple[ActionBatch, dict]:
        raise NotImplementedError

    def reset_state(self) -> None:
        pass

    @abstractmethod
    def train_step(self, batch_data: BatchSequenceData, step: int) -> int:
        """Returns the new step count (sometimes multiple gradient steps are taken here)."""
        raise NotImplementedError



from typing import Optional

import attr
import torch

from avalon.agent.common.models import ActivationFunction
from avalon.agent.common.params import Params


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class ObservationModelParams:
    encoder_output_dim: int = 256
    num_mlp_layers: int = 2
    num_cnn_base_channels: int = 16
    mlp_activation_fn: ActivationFunction = ActivationFunction.ACTIVATION_RELU


@attr.s(auto_attribs=True, frozen=True)
class OnPolicyParams(Params):
    num_steps: int = 256

    @property
    def inference_device(self):
        return torch.device(f"cuda:{self.inference_gpus[0]}")


@attr.s(auto_attribs=True, frozen=True)
class PPOParams(OnPolicyParams):
    # Default overrides
    project: str = "avalon_ppo"
    # don't routinely change these!
    obs_first: bool = True
    time_limit_bootstrapping: bool = True
    num_steps: int = 256
    num_workers: int = 8
    batch_size: int = 256 * 8

    # PPO-specific stuff
    model: str = "godot"
    discount: float = 0.99
    lam: float = 0.95
    clip_range: float = 0.2
    clip_range_vf: Optional[float] = None
    value_loss_coef: float = 0.5
    baselines_style_vf_loss: bool = False
    entropy_mode: str = "regularized"  # can be "max" or "regularized", same meaning as in `garage` VPG
    entropy_coef: float = 0.0  # entropy coef for use in either entropy mode. set to 0 to disable both.

    ppo_epochs: int = 10
    lr: float = 3e-4
    clip_grad_norm: float = 0.5

    normal_std_from_model: bool = True
    # policy_normal_init_std: float = 1
    # policy_normal_min_std: float = 0.01
    # clipped_normal_mode: ClippedNormalMode = ClippedNormalMode.TRUNCATED_NORMAL

    model_params: ObservationModelParams = ObservationModelParams()

    def __attrs_post_init__(self) -> None:
        assert len(self.inference_gpus) == 1, "we can only use one inference gpu"

        # These should only be changed with great care.
        assert self.env_params.action_repeat == 1
        assert self.obs_first is True

    @property
    def num_batches(self) -> int:
        assert (
            self.num_steps * self.num_workers % self.batch_size == 0
        ), f"batch size {self.batch_size} does not divide cleanly into steps ({self.num_steps}) * workers ({self.num_workers})"
        return int(self.num_steps * self.num_workers / self.batch_size)



from enum import Enum
from typing import Callable
from typing import Dict
from typing import List

import torch
from torch import Tensor


class ActivationFunction(Enum):
    ACTIVATION_EXP = "ACTIVATION_EXP"
    ACTIVATION_SOFTPLUS = "ACTIVATION_SOFTPLUS"
    ACTIVATION_ELU = "ACTIVATION_ELU"
    ACTIVATION_RELU = "ACTIVATION_RELU"
    ACTIVATION_LEAKY_RELU = "ACTIVATION_LEAKY_RELU"
    ACTIVATION_TANH = "ACTIVATION_TANH"


# maps from str instead of function so this even works with reloading
ACTIVATION_MODULE_LOOKUP: Dict[str, Callable[[], torch.nn.Module]] = {
    ActivationFunction.ACTIVATION_SOFTPLUS.value: torch.nn.Softplus,
    ActivationFunction.ACTIVATION_ELU.value: torch.nn.ELU,
    ActivationFunction.ACTIVATION_RELU.value: torch.nn.ReLU,
    ActivationFunction.ACTIVATION_LEAKY_RELU.value: torch.nn.LeakyReLU,
    ActivationFunction.ACTIVATION_TANH.value: torch.nn.Tanh,
}


class MLP(torch.nn.Module):
    def __init__(
        self,
        input_dim: int,
        hidden_dim: int,
        output_dim: int,
        num_layers: int,
        activation_function: ActivationFunction = ActivationFunction.ACTIVATION_RELU,
        is_batch_normalized: bool = False,
    ) -> None:
        """Will have `num_layers` `torch.nn.Linear` layers.

        So num_layers=2 will yield a MLP with one hidden layer and one output layer."""
        super().__init__()
        self.act_fn = ACTIVATION_MODULE_LOOKUP[activation_function.value]

        assert num_layers >= 0, "negative layers?!?"
        if num_layers == 0:
            assert input_dim == output_dim, "For identity MLP, input dim must equal output dim"
            self.net: torch.nn.Module = torch.nn.Identity()
            return

        if num_layers == 1:
            self.net = torch.nn.Linear(input_dim, output_dim)
            return

        layers: List[torch.nn.Module] = []
        prev_dim = input_dim
        for _ in range(num_layers - 1):
            layers.append(torch.nn.Linear(prev_dim, hidden_dim))
            layers.append(self.act_fn())
            if is_batch_normalized:
                layers.append(torch.nn.BatchNorm1d(hidden_dim))

            prev_dim = hidden_dim

        layers.append(torch.nn.Linear(hidden_dim, output_dim))

        self.net = torch.nn.Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        result = self.net(x)  # type: ignore
        assert isinstance(result, Tensor)
        return result



from enum import Enum
from typing import Optional

import attr
import gym.spaces
import torch


class ClippedNormalMode(Enum):
    NO_CLIPPING = "NO_CLIPPING"
    SAMPLE_DIST = "SAMPLE_DIST"
    TRUNCATED_NORMAL = "TRUNCATED_NORMAL"


@attr.s(auto_attribs=True, frozen=True)
class EnvironmentParams:
    suite: str = "godot"
    task: Optional[str] = None
    env_index: int = 0  # each environment will get a unique worker id in range [0, num_workers * num_worker_groups)
    env_count: int = 1  # this will get set automatically to num_workers
    action_repeat: int = 1
    time_limit: Optional[int] = None
    reward_scale: float = 1  # scale reward magnitude by this amount
    pixel_obs_wrapper: bool = False
    elapsed_time_obs: bool = False  # include an observation of the elapsed time
    mode: str = "train"
    frame_stack: int = 1
    info_fields: list[str] = [
        "cumulative_episode_return",
        "cumulative_episode_length",
    ]  # only non-nested scalar float-castable supported


@attr.s(auto_attribs=True, frozen=True)
class DmcEnvironmentParams(EnvironmentParams):
    suite: str = "dmc"
    include_proprio: bool = True
    include_rgb: bool = False


@attr.s(auto_attribs=True, frozen=True)
class ProcgenEnvironmentParams(EnvironmentParams):
    suite: str = "procgen"
    task: str = "starpilot"
    num_levels: int = 0
    start_level: int = 0
    distribution_mode: str = "easy"


@attr.s(auto_attribs=True, frozen=True)
class Params:
    # wandb
    project: str
    name: Optional[str] = None  # wandb run name
    tags: tuple[str, ...] = ()
    wandb_mode: str = "online"  # online, offline, disabled
    log_freq_hist: int = 50
    log_freq_scalar: int = 1
    log_freq_media: int = 250
    group: Optional[str] = None
    suggestion_uuid: str = ""  # for use by BONES and friends

    # environment
    env_params: EnvironmentParams = EnvironmentParams()

    # worker
    multiprocessing: bool = True
    num_workers: int = 8  # number of environments per worker group
    num_steps: int = 256  # number of steps per rollout
    time_limit_bootstrapping: bool = False
    # if obs_first=True, for a given timestep, the order is (observation, action, reward/done/info)
    # if obs_first=False, for a given timestep, the order is (action, reward/done/info, observation)
    obs_first: bool = True
    center_observations: bool = True
    allow_partial_batches: bool = True  # should we wait for all envs to be ready, or allow a partial batch?

    train_gpu: int = 0  # currently this does nothing (but will happen automatically)
    inference_gpus: tuple[int, ...] = (0,)  # only applied if there are async worker managers
    multiprocessing_mode: str = "fork"  # `spawn` or `fork`. `fork` is faster, `spawn` more compatable

    # training
    is_training: bool = True
    batch_size: int = 100
    # a string specifying the possible places to resume from, either local file or wandb run
    # valid formats include:
    #     wandb://{project}/{run}/{file_name}
    #     file://{absolute_file_path}
    # for example:
    #     wandb://untitled-ai/sf3189ytcfg/checkpoint.pt
    #     file:///home/user/runs/good_run/checkpoint.pt
    resume_from: Optional[str] = None
    checkpoint_every: int = 10_000
    total_env_steps: int = 1_000_000
    discount = 0.98

    # Policy/action head parameters
    normal_std_from_model: bool = True
    policy_normal_init_std: float = 1
    policy_normal_min_std: float = 0.01
    clipped_normal_mode: ClippedNormalMode = ClippedNormalMode.NO_CLIPPING  # is this really the best default now?
    center_and_clamp_discrete_logits: bool = False

    # val/test
    is_testing: bool = False
    eval_exploration_mode: str = "eval"  # "eval" to use the mode of the distribution, "explore" to sample
    eval_workers: int = 16

    # spaces
    observation_space: Optional[gym.spaces.Dict] = None
    action_space: Optional[gym.spaces.Dict] = None

    debug: bool = False  # Disables some crash-recovery that obscures crashes when debugging.
    deterministic: bool = False

    @property
    def train_device(self) -> torch.device:
        return torch.device(f"cuda:{self.train_gpu}")



import queue
import time
import uuid
from functools import partial
from pathlib import Path
from typing import Iterator

import attr
import torch
from loguru import logger
from torch.utils.data import DataLoader

from avalon.agent.common import wandb_lib
from avalon.agent.common.dataloader import ReplayDataset
from avalon.agent.common.dataloader import worker_init_fn
from avalon.agent.common.params import DmcEnvironmentParams
from avalon.agent.common.parse_args import parse_args
from avalon.agent.common.storage import EpisodeStorage

# from avalon.agent.common.storage import StorageMode
from avalon.agent.common.trainer import Trainer
from avalon.agent.common.types import BatchSequenceData
from avalon.agent.common.util import pack_1d_list
from avalon.agent.common.util import setup_new_process
from avalon.agent.common.worker import RolloutManager
from avalon.agent.dreamer.params import DreamerParams
from avalon.agent.dreamer.params import OffPolicyParams
from avalon.common.log_utils import configure_remote_logger


class DreamerTrainer(Trainer[OffPolicyParams]):
    """This `Trainer` is designed to exactly replicate the training setup in danijar's dreamerv2 repo.

    Instead of running the data-collection and training in separate, unsynchronized processes, as we do normally,
    this runs them in an alternating manner in a single process.
    This results in a fixed constant ratio between training and env steps.
    """

    def __init__(self, params: OffPolicyParams) -> None:
        self.wandb_queue = queue.Queue()  # type: ignore[var-annotated]
        self.train_rollout_dir = str(Path(params.data_dir) / "train" / str(uuid.uuid4()))
        super().__init__(params)

        # Prefill so we have enough steps to form a first batch
        rollout_steps = params.prefill_steps // params.num_workers
        self.train_rollout_manager.run_rollout(
            num_steps=rollout_steps,
            num_episodes=1,
            exploration_mode="explore",
        )
        self.env_step += self.params.num_workers * rollout_steps

        # Pretrain (train without doing fresh rollouts). Lets us train a bit before using the model for rollouts.
        if not self.start:
            self.start = time.time()
        for _ in range(params.pretrain_steps):
            super().train_step()

    def create_rollout_manager(self):
        rollout_manager = RolloutManager(
            params=self.params,
            num_workers=self.params.num_workers,
            is_multiprocessing=self.params.multiprocessing,
            storage=self.train_storage,
            obs_space=self.params.observation_space,
            model=self.algorithm,
            rollout_device=torch.device(f"cuda:{self.params.inference_gpus[0]}"),
            multiprocessing_context=self.multiprocessing_context,
        )
        self.to_cleanup.append(rollout_manager)
        return rollout_manager

    def create_train_storage(self) -> EpisodeStorage:
        return EpisodeStorage(self.params, self.train_rollout_dir, self.wandb_queue)

    def create_dataloader(self) -> Iterator[BatchSequenceData]:
        train_dataset = ReplayDataset(self.params, self.train_rollout_dir, update_interval=4000)
        return iter(
            DataLoader(
                train_dataset,
                batch_size=self.params.batch_size,
                shuffle=False,
                num_workers=self.params.num_dataloader_workers,
                drop_last=True,
                pin_memory=True,
                prefetch_factor=2,
                worker_init_fn=worker_init_fn,
                collate_fn=partial(pack_1d_list, out_cls=BatchSequenceData),
            )
        )

    def train_step(self) -> None:
        rollout_steps = self.params.train_every
        self.train_rollout_manager.run_rollout(
            num_steps=rollout_steps,
            exploration_mode="explore",
        )
        self.env_step += self.params.num_workers * rollout_steps

        old_i = self.i
        super().train_step()
        assert self.i == old_i + 1, "Off-policy algorithms must increment i by only 1"

        wandb_lib.log_from_queue(self.wandb_queue, prefix=f"rollout/")

    @property
    def frames_per_batch(self):
        assert self.params.min_fragment_len == self.params.max_fragment_len
        return self.params.batch_size * self.params.min_fragment_len


@attr.s(auto_attribs=True, frozen=True)
class DreamerDMCParams(DreamerParams):
    name: str = "torch"
    total_env_steps: int = 250_000
    pcont: bool = False
    prioritize_ends: bool = False
    worker_managers: int = 1
    num_workers: int = 1
    env_params: DmcEnvironmentParams = DmcEnvironmentParams(
        suite="dmc", task="walker_walk", action_repeat=2, time_limit=None
    )
    free_nats: float = 1
    policy_entropy_scale: float = 1e-4
    model_lr: float = 3e-4
    actor_lr: float = 8e-5
    value_lr: float = 8e-5
    deter_size: int = 200
    rssm_hidden_size: int = 200
    train_gpu: int = 0
    inference_gpus: tuple[int, ...] = (0,)
    log_freq_hist: int = 100
    log_freq_scalar: int = 20
    log_freq_media: int = 500
    checkpoint_every: int = 2500
    num_dataloader_workers: int = 1
    log_rollout_metrics_every: int = 1
    prefill_steps: int = 1000
    pretrain_steps: int = 100
    freeze_actor_steps: int = 0
    prefill_eps_per_dataloader: int = 1


def run(params: DreamerDMCParams) -> None:
    # TODO: do this!!
    # assert False, "check param counts match before running this again!"

    trainer = DreamerTrainer(params)
    try:
        trainer.train()
    finally:
        trainer.shutdown()


if __name__ == "__main__":
    configure_remote_logger(level="INFO")
    setup_new_process()
    try:
        default_params = DreamerDMCParams()
        default_params = parse_args(default_params)
        # default_params = attr.evolve(
        #     default_params,
        #     name=f"{default_params.name}_{default_params.env_params.suite}_{default_params.env_params.task}",
        # )
        run(default_params)
    except Exception as e:
        logger.exception(e)
        raise



import argparse
import os
import time
import warnings
from functools import partial
from pathlib import Path

import attr
import torch
from loguru import logger
from pytest import MonkeyPatch
from requests.exceptions import RequestsDependencyWarning

from avalon.agent.common.envs import build_env
from avalon.agent.common.params import EnvironmentParams
from avalon.agent.common.params import Params
from avalon.agent.common.trainer import OffPolicyTrainer
from avalon.agent.common.trainer import OnPolicyTrainer
from avalon.agent.common.trainer import Trainer
from avalon.agent.common.util import hash_model
from avalon.agent.common.util import setup_new_process
from avalon.agent.dreamer.params import DreamerParams
from avalon.agent.ppo.params import PPOParams
from avalon.agent.tests.fixtures import _make_agent
from avalon.agent.tests.fixtures import complete_agent_params
from avalon.agent.tests.fixtures import get_valid_agent_env_combos
from avalon.agent.tests.utils import _dump
from avalon.agent.tests.utils import _load
from avalon.agent.tests.utils import run_deterministic_forward_pass
from avalon.common.log_utils import configure_remote_logger
from avalon.contrib.utils import make_deterministic
from avalon.datagen.world_creation.tests.params import CANONICAL_SEED


def generate_test_input_data(seed: int, test_data_path: Path) -> None:
    for (agent_name, env_name), (agent_params, env_params) in get_valid_agent_env_combos(seed).items():
        logger.info(f"Generating input data for {(agent_name, env_name)}")
        make_deterministic(seed)
        env = build_env(env_params)
        env.action_space.seed(seed)
        initial_observation = env.reset()
        _dump(initial_observation, test_data_path / f"{env_name}_initial_observation.pkl")
        random_observation, *rest = env.step(env.action_space.sample())
        _dump(random_observation, test_data_path / f"{env_name}_random_observation.pkl")
        env.close()
        del env


def run_deterministic_train_step(
    agent_params: Params, env_params: EnvironmentParams, seed: int = CANONICAL_SEED
) -> Trainer:
    with MonkeyPatch.context() as mp:
        # TODO: d67559db-84db-4d5f-8392-d551ed6a75c5
        # Deterministic backward passes on GPU need a deterministic scatter_add implementation that's only on Torch 1.13
        # see: https://github.com/pytorch/pytorch/commit/5b58140d1a471b144baf66cc61a45a86746f0215
        # Note that inference (or any other model in a different process) will still run on GPU - we only care about the
        # learner since that's where `scatter_add` is used.
        cpu_device = torch.device("cpu")

        def forced_cpu(_device: str) -> torch.device:
            return cpu_device

        mp.setattr(torch, "device", forced_cpu)
        mp.setattr(torch.cuda, "is_available", lambda: False)

        # We can't monkeypatch the functions directly to seed, since some of them are in different processes
        mp.setenv("AVALON_MODEL_SEED", str(seed))

        if isinstance(agent_params, DreamerParams):
            # For OffPolicyTrainer, we want to sleep as little as possible so that we load the first episode ASAP,
            # otherwise we risk non-determinism if two eps are there when we load in data (in which case their order is
            # non-deterministic since we can't trivially patch uuids to be non-random (as they use os.urandom)
            sleep_100ms = partial(time.sleep, 0.1)

            def quick_sleep(_secs: int) -> None:
                sleep_100ms()

            mp.setattr(time, "sleep", quick_sleep)

            agent_params = complete_agent_params(agent_params, env_params, seed=seed)
            agent_params = attr.evolve(agent_params, min_fragment_len=30, max_fragment_len=30, batch_size=10)
            make_deterministic(seed)
            dreamer_trainer = OffPolicyTrainer(agent_params)
            dreamer_trainer.start = time.time()
            try:
                dreamer_trainer.start_train_rollouts()
                dreamer_trainer.train_step()
            finally:
                dreamer_trainer.shutdown()
            return dreamer_trainer
        elif isinstance(agent_params, PPOParams):
            agent_params = complete_agent_params(agent_params, env_params, seed=seed)
            agent_params = attr.evolve(agent_params, batch_size=8, num_steps=32)
            ppo_trainer = OnPolicyTrainer(agent_params)
            ppo_trainer.start = time.time()
            try:
                ppo_trainer.train_step()
            finally:
                ppo_trainer.shutdown()
            return ppo_trainer
        else:
            raise NotImplementedError(type(agent_params))


def generate_test_output_data(seed: int, test_data_path: Path) -> None:
    for (agent_name, env_name), (agent_params, env_params) in get_valid_agent_env_combos(seed).items():
        logger.info(f"Generating output data for {(agent_name, env_name)}")
        for observation_kind in ("initial", "random"):
            observation = _load(test_data_path / f"{env_name}_{observation_kind}_observation.pkl")
            agent = _make_agent(complete_agent_params(agent_params, env_params, seed=seed))
            output = run_deterministic_forward_pass(agent, observation, seed)
            _dump(output, test_data_path / f"{agent_name}_{env_name}_{observation_kind}_forward_pass.pkl")
            del agent

        trainer = run_deterministic_train_step(agent_params, env_params, seed)
        model_hash = hash_model(trainer.algorithm)
        logger.info(f"{model_hash=}")
        _dump(model_hash, test_data_path / f"{agent_name}_{env_name}_trained_model_hash.pkl")


if __name__ == "__main__":
    configure_remote_logger(level="INFO")
    setup_new_process()
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    warnings.filterwarnings("ignore", category=RequestsDependencyWarning)

    parser = argparse.ArgumentParser()
    parser.add_argument("--input", action="store_const", const=True)
    parser.add_argument("--output", action="store_const", const=True)
    parser.add_argument("--seed", action="store", type=int, default=CANONICAL_SEED)
    args = parser.parse_args()

    os.environ["WANDB_MODE"] = "disabled"
    output_path = Path(__file__).parent / "data"
    output_path.mkdir(exist_ok=True, parents=True)
    assert args.input or args.output, "Must specify which test data to generate"
    if args.input:
        generate_test_input_data(args.seed, test_data_path=output_path)
    if args.output:
        generate_test_output_data(args.seed, test_data_path=output_path)



from pathlib import Path
from typing import Callable
from typing import Dict
from typing import Tuple

import attr
import numpy as np
from scipy import stats
from skimage import morphology
from skimage.morphology import flood_fill

from avalon.common.utils import only
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.task import TaskConfig
from avalon.datagen.world_creation.constants import ITEM_FLATTEN_RADIUS
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.entities.food import CANONICAL_FOOD_CLASS
from avalon.datagen.world_creation.entities.tools.log import Log
from avalon.datagen.world_creation.entities.tools.stone import Stone
from avalon.datagen.world_creation.entities.tools.weapons import LargeRock
from avalon.datagen.world_creation.entities.tools.weapons import Rock
from avalon.datagen.world_creation.entities.tools.weapons import Stick
from avalon.datagen.world_creation.tasks.bridge import create_bridge_obstacle
from avalon.datagen.world_creation.tasks.eat import add_food_tree_for_simple_task
from avalon.datagen.world_creation.tasks.fight import create_fight_obstacle
from avalon.datagen.world_creation.tasks.stack import create_stack_obstacle
from avalon.datagen.world_creation.tasks.stack import flatten_places_under_items
from avalon.datagen.world_creation.tasks.throw import throw_obstacle_maker
from avalon.datagen.world_creation.utils import to_2d_point
from avalon.datagen.world_creation.worlds.difficulty import scale_with_difficulty
from avalon.datagen.world_creation.worlds.export import export_world


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class CarryTaskConfig(TaskConfig):
    # how far should the item be moved away from its original spot on difficulty 0.0 and 1.0 respectively
    carry_dist_easy: float = 1.0
    carry_dist_hard: float = 10.0
    # in which tasks should we be carrying items away from their original spawn locations?
    # the way the carry task works is simply by moving the items farther away from their original locations
    tasks: Tuple[AvalonTask, ...] = (AvalonTask.THROW, AvalonTask.FIGHT, AvalonTask.STACK, AvalonTask.BRIDGE)


_GENERATION_FUNCTION_BY_TASK: Dict[AvalonTask, Callable] = {
    AvalonTask.THROW: throw_obstacle_maker,
    AvalonTask.FIGHT: create_fight_obstacle,
    AvalonTask.STACK: create_stack_obstacle,
    AvalonTask.BRIDGE: create_bridge_obstacle,
}


def generate_carry_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: CarryTaskConfig = CarryTaskConfig(),
) -> None:
    inner_task = rand.choice(task_config.tasks)  # type: ignore[arg-type]
    inner_generator = _GENERATION_FUNCTION_BY_TASK[inner_task]
    result = inner_generator(rand, difficulty, export_config, is_for_carry=True)
    if len(result) == 5:
        difficulty, food_class, locations, spawn_location, world = result
        locations = attr.evolve(locations, spawn=spawn_location)
        is_on_tree = False
    else:
        world, locations, difficulty = result
        food_class = CANONICAL_FOOD_CLASS
        is_on_tree = True

    world, locations = world.end_height_obstacles(locations, is_accessible_from_water=False)

    # move the things to your spawn region

    # start with this island
    mask = locations.island
    # remove all obstacles
    mask = np.logical_and(mask, np.logical_not(world.full_obstacle_mask))
    # remove unclimbable places
    mask = np.logical_and(mask, world.is_climbable)
    # make a little smaller:
    mask = morphology.dilation(np.logical_not(mask), morphology.disk(1))

    # plot_value_grid(mask)

    # flood fill from your spanw. These are the points that you can reach
    index = world.map.point_to_index(to_2d_point(locations.spawn))
    flood_mask = mask.copy().astype(int)
    flood_fill(flood_mask, index, 2, in_place=True)
    spawn_region = flood_mask == 2

    # plot_value_grid(spawn_region, "spawn")

    if np.any(spawn_region):
        all_mobile_items = [x for x in world.items if getattr(x, "solution_mask", None) is not None]
        for item in all_mobile_items:
            full_position_mask = np.logical_or(item.solution_mask, spawn_region)
            item = attr.evolve(item, solution_mask=full_position_mask)
            desired_distance = scale_with_difficulty(
                difficulty, task_config.carry_dist_easy, task_config.carry_dist_hard
            )
            carry_distance = stats.norm(desired_distance, desired_distance / 4.0)
            world = world.carry_tool_randomly(rand, item, carry_distance)

    if inner_task == AvalonTask.STACK:
        world = flatten_places_under_items(world, earlier_item_count=0, filter_to_classes=Stone)
    elif inner_task == AvalonTask.BRIDGE:
        log = only([x for x in world.items if isinstance(x, Log)])
        world = world.flatten(to_2d_point(log.position), ITEM_FLATTEN_RADIUS, ITEM_FLATTEN_RADIUS)
    else:
        world = flatten_places_under_items(world, earlier_item_count=0, filter_to_classes=(Rock, LargeRock, Stick))

    if is_on_tree:
        world = add_food_tree_for_simple_task(world, locations)
        world = world.add_spawn(rand, difficulty, locations.spawn, locations.goal)
    else:
        world = world.add_spawn_and_food(rand, difficulty, locations.spawn, locations.goal, food_class=food_class)

    export_world(output_path, rand, world)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Naive profiling using timeit. (Used in MonoBeast.)"""

import collections
import timeit


class Timings:
    """Not thread-safe."""

    def __init__(self) -> None:
        self._means = collections.defaultdict(int)
        self._vars = collections.defaultdict(int)
        self._counts = collections.defaultdict(int)
        self.reset()

    def reset(self) -> None:
        self.last_time = timeit.default_timer()

    def time(self, name: str) -> None:
        """Save an update for event `name`.

        Nerd alarm: We could just store a
            collections.defaultdict(list)
        and compute means and standard deviations at the end. But thanks to the
        clever math in Sutton-Barto
        (http://www.incompleteideas.net/book/first/ebook/node19.html) and
        https://math.stackexchange.com/a/103025/5051 we can update both the
        means and the stds online. O(1) FTW!
        """
        now = timeit.default_timer()
        x = now - self.last_time
        self.last_time = now

        n = self._counts[name]

        mean = self._means[name] + (x - self._means[name]) / (n + 1)
        var = (n * self._vars[name] + n * (self._means[name] - mean) ** 2 + (x - mean) ** 2) / (n + 1)

        self._means[name] = mean
        self._vars[name] = var
        self._counts[name] += 1

    def means(self):
        return self._means

    def vars(self):
        return self._vars

    def stds(self):
        return {k: v**0.5 for k, v in self._vars.items()}

    def summary(self, prefix: str = "") -> str:
        means = self.means()
        stds = self.stds()
        total = sum(means.values())

        result = prefix
        for k in sorted(means, key=means.get, reverse=True):
            result += f"\n    %s: %.6fms +- %.6fms (%.2f%%) " % (
                k,
                1000 * means[k],
                1000 * stds[k],
                100 * means[k] / total,
            )
        result += "\nTotal: %.6fms" % (1000 * total)
        return result



from typing import Optional

import attr

from avalon.datagen.world_creation.indoor.builders import EntranceBuilder
from avalon.datagen.world_creation.indoor.builders import FootprintBuilder
from avalon.datagen.world_creation.indoor.builders import HallwayBuilder
from avalon.datagen.world_creation.indoor.builders import RoomBuilder
from avalon.datagen.world_creation.indoor.builders import StoryLinker
from avalon.datagen.world_creation.indoor.builders import WindowBuilder
from avalon.datagen.world_creation.indoor.building import BuildingAestheticsConfig


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class BuildingConfig:
    width: int  # x
    length: int  # z
    story_count: int
    footprint_builder: FootprintBuilder
    room_builder: RoomBuilder
    hallway_builder: HallwayBuilder
    story_linker: Optional[StoryLinker] = None
    entrance_builder: Optional[EntranceBuilder] = None
    window_builder: Optional[WindowBuilder] = None
    is_climbable: bool = True
    aesthetics: BuildingAestheticsConfig = BuildingAestheticsConfig()


def generate_aesthetics_config() -> BuildingAestheticsConfig:
    return BuildingAestheticsConfig()



#!/usr/bin/env python
import os
import stat
import sys
import urllib.request
from pathlib import Path
from typing import Dict
from typing import Final
from typing import List
from typing import Literal
from typing import Tuple
from zipfile import ZipFile

from loguru import logger

from avalon.datagen.godot_env.interactive_godot_process import GODOT_BINARY_PATH
from avalon.datagen.godot_env.interactive_godot_process import GODOT_EDITOR_PATH

RELEASES: Final = "https://github.com/Avalon-Benchmark/godot/releases/download"

CURRENT_RELEASE: Final = "3.4.4.avalon.0.9.3"

Platform = Literal["linux", "macos", "windows"]

BinaryType = Literal["runner", "editor"]

RUNNER: Final = "runner"
EDITOR: Final = "editor"

CWD: Final = os.getcwd()


def _friendly_path(path: Path) -> str:
    abs_path = str(path.resolve())
    try:
        relative = str(path.relative_to(CWD))
        if len(relative) < len(abs_path):
            return relative
    except ValueError:
        pass
    return abs_path


def get_platform() -> Platform:
    p = sys.platform
    if p == "linux":
        return "linux"
    if p == "darwin":
        return "macos"
    if p in ("win32", "cygwin"):
        return "windows"
    raise ValueError(f"No avalon godot binaries available for {p}")


builds: Final[Dict[Tuple[Platform, BinaryType], str]] = {
    ("linux", RUNNER): "linux-egl-editor.zip",
    ("linux", EDITOR): "linux-x11-editor.zip",
    ("macos", EDITOR): "macos-editor.zip",
    ("windows", EDITOR): "windows-editor.zip",
}


def available_builds() -> List[BinaryType]:
    this_platform = get_platform()
    return [build for plat, build in builds.keys() if plat == this_platform]


def get_godot_binary_path() -> Path:
    path = Path(GODOT_BINARY_PATH)
    path.parent.mkdir(exist_ok=True)
    return path


def get_godot_editor_path() -> Path:
    path = Path(GODOT_EDITOR_PATH)
    path.parent.mkdir(exist_ok=True)
    return path


def _extract_singular_as(zipped: Path, target: Path) -> None:
    with ZipFile(zipped, "r") as archive:
        members = archive.namelist()
        assert len(members) == 1, "Should only contain a single executable binary"
        archive.extract(members[0], target.parent)
        extracted_destination = target.parent / members[0]
    os.rename(extracted_destination, target)


def fetch_binary(build: str, target: Path) -> None:
    archive_path = target.parent / build
    release = f"{CURRENT_RELEASE}/{build}"
    logger.info(f"Downloading {release} into {_friendly_path(target)}")
    urllib.request.urlretrieve(f"{RELEASES}/{release}", archive_path)
    _extract_singular_as(archive_path, target)
    os.remove(archive_path)


def ensure_executable(path: Path) -> None:
    current_permissions_plus_executable = os.stat(path).st_mode | stat.S_IEXEC
    os.chmod(path, current_permissions_plus_executable)


def handle_overwrite(target_path: Path, is_overwriting: bool) -> None:
    if target_path.exists() or target_path.is_symlink():
        if not is_overwriting:
            logger.error(f"Refusing to overwrite existing file {target_path} without --overwrite")
            exit(1)
        os.remove(target_path)


def install_binary(build: str, target_path: Path, is_overwriting: bool) -> None:
    handle_overwrite(target_path, is_overwriting)
    if target_path.exists():
        if not is_overwriting:
            logger.error(f"Refusing to overwrite existing file {target_path} without --overwrite")
            exit(1)
        os.remove(target_path)
    fetch_binary(build, target_path)
    ensure_executable(target_path)


def install_available_binaries_for_current_platform(is_overwriting: bool) -> None:
    platform = get_platform()
    binary_path = get_godot_binary_path()
    editor_path = get_godot_editor_path()

    editor_build = builds.get((platform, EDITOR), None)
    headless_build = builds.get((platform, RUNNER), None)
    assert editor_build is not None, f"{platform} is currently unsupported"
    logger.info(
        f"Installing editor {'build' if headless_build is None else 'and headless runner builds'} from {RELEASES}"
    )

    install_binary(editor_build, editor_path, is_overwriting)

    if headless_build is not None:
        install_binary(headless_build, binary_path, is_overwriting)
    elif platform == "windows":
        logger.info(
            f"Note: Installed godot editor for windows, but it cannot be used as a runner. "
            f"Consider using WSL for actual training purposes: https://learn.microsoft.com/en-us/windows/wsl/install"
        )
    else:
        logger.info(
            f"Note: No headless runner currently available for {platform}. Symlinking editor build in it's place."
        )
        handle_overwrite(binary_path, is_overwriting)
        os.symlink(editor_path, binary_path, is_overwriting)


if __name__ == "__main__":
    is_overwriting = len(sys.argv) > 1 and sys.argv[1] == "--overwrite"
    install_available_binaries_for_current_platform(is_overwriting)



from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import cast

import attr
import numpy as np
from IPython.display import HTML
from IPython.display import display
from matplotlib import pyplot as plt
from scipy.interpolate import RegularGridInterpolator
from scipy.spatial.transform import Rotation
from skimage import morphology

from avalon.common.errors import SwitchError
from avalon.common.utils import only
from avalon.datagen.world_creation.configs.biome import BiomeConfig
from avalon.datagen.world_creation.configs.biome import _raw_biome_id_mapping
from avalon.datagen.world_creation.configs.scenery import SceneryConfig
from avalon.datagen.world_creation.constants import DARK_SHORE
from avalon.datagen.world_creation.constants import FRESH_WATER
from avalon.datagen.world_creation.constants import MAX_CLIFF_TERRAIN_ANGLE
from avalon.datagen.world_creation.constants import MIN_CLIFF_TERRAIN_ANGLE
from avalon.datagen.world_creation.constants import SWAMP
from avalon.datagen.world_creation.constants import UNCLIMBABLE_BIOME_ID
from avalon.datagen.world_creation.constants import UP_VECTOR
from avalon.datagen.world_creation.constants import WATER_LINE
from avalon.datagen.world_creation.debug_plots import plot_value_grid
from avalon.datagen.world_creation.entities.scenery import Scenery
from avalon.datagen.world_creation.noise import perlin
from avalon.datagen.world_creation.types import Biome
from avalon.datagen.world_creation.types import FloatListNP
from avalon.datagen.world_creation.types import MapBoolNP
from avalon.datagen.world_creation.types import MapFloatNP
from avalon.datagen.world_creation.types import MapIntNP
from avalon.datagen.world_creation.types import Point2DListNP
from avalon.datagen.world_creation.types import Point3DListNP
from avalon.datagen.world_creation.types import SceneryBorderMode
from avalon.datagen.world_creation.utils import hex_to_rgb
from avalon.datagen.world_creation.worlds.height_map import HeightMap
from avalon.datagen.world_creation.worlds.height_map import SpecialBiomes
from avalon.datagen.world_creation.worlds.height_map import create_kd_tree
from avalon.datagen.world_creation.worlds.height_map import get_border_points
from avalon.datagen.world_creation.worlds.height_map import get_flora_config_by_file


def sidewaysness(normals: Point3DListNP) -> FloatListNP:
    xzlen = np.sqrt(normals[:, 0] ** 2 + normals[:, 2] ** 2)
    angle = np.abs(np.arctan2(normals[:, 1], xzlen))
    result = np.clip(((MAX_CLIFF_TERRAIN_ANGLE - angle)) / (MAX_CLIFF_TERRAIN_ANGLE - MIN_CLIFF_TERRAIN_ANGLE), 0, 1)
    result[angle < MIN_CLIFF_TERRAIN_ANGLE] = 1.0
    return cast(FloatListNP, result)


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class BiomeMap:
    map: HeightMap
    biomes: Tuple[Biome, ...]
    biome_id: MapIntNP
    border_distances: MapFloatNP
    config: BiomeConfig

    def get_colors(
        self, rand: np.random.Generator, is_climbable: MapBoolNP, points: Point3DListNP, normals: Point3DListNP
    ) -> np.ndarray:
        grid_x, grid_y = self.map.region.get_linspace(self.map.cells_per_meter)
        biome_id_with_unclimb = self.biome_id.copy()
        is_steep = self.map.get_squared_slope() > self.config.high_slope_threshold
        biome_id_with_unclimb[np.logical_and(is_steep, np.logical_not(is_climbable))] = UNCLIMBABLE_BIOME_ID
        interp_spline = RegularGridInterpolator((grid_x, grid_y), biome_id_with_unclimb, method="nearest")
        points_2d = np.stack([points[:, 2], points[:, 0]], axis=1)
        biome_ids = interp_spline(points_2d)

        color_mapping = self.config.color_map
        neighbor_count = self.config.color_sampling_neighbor_count
        if neighbor_count == 0:
            colors = np.array([color_mapping[x] for x in biome_ids])
        else:
            min_val = self.map.region.x.min_ge
            max_val = self.map.region.x.max_lt
            neighbors_2d = points_2d.reshape((len(points_2d), 1, 2)).repeat(neighbor_count, 1)
            neighbors_2d += rand.standard_normal(neighbors_2d.shape) * self.config.color_sampling_std_dev
            neighbors_2d = neighbors_2d.reshape((len(neighbors_2d) * neighbor_count, 2))
            neighbors_2d = np.clip(neighbors_2d, min_val, max_val)
            neighbor_biome_ids = interp_spline(neighbors_2d)
            colors = np.array([color_mapping[x] for x in neighbor_biome_ids])
            colors = colors.reshape((len(points), neighbor_count, 3)).mean(axis=1)

        # coast is special
        is_coast = biome_ids == self.config.coastal_id
        colors[is_coast] = (
            np.array(color_mapping[self.config.coastal_id]) * (1.0 - self.config.coast_color_noise)
            + self.config.coast_color_noise * colors[is_coast]
        )

        # rocks are special too
        for rock_id in self.config.rock_biome_ids:
            is_rock = biome_ids == rock_id
            colors[is_rock] = (
                np.array(color_mapping[rock_id]) * (1.0 - self.config.rock_color_noise)
                + self.config.rock_color_noise * colors[is_rock]
            )

        unclimbable_color = color_mapping[UNCLIMBABLE_BIOME_ID]
        climb_spline = RegularGridInterpolator((grid_x, grid_y), is_climbable)
        is_climbable_factor = climb_spline(points_2d)
        unclimbability = ((1.0 - is_climbable_factor) * sidewaysness(normals)).reshape(-1, 1)
        colors = unclimbable_color * unclimbability + colors * (1.0 - unclimbability)

        # apply correlated color noise
        d = {x.id: x.correlated_color_jitter for x in self.biomes}
        u, inv = np.unique(biome_ids, return_inverse=True)
        jitters = np.array([d[x] for x in u])[inv].reshape(biome_ids.shape)
        jitters = jitters.reshape(*jitters.shape, 1)
        colors = np.clip(colors * 1.0 + rand.normal(0.0, jitters, jitters.shape).repeat(3, axis=-1), 0.0, 1.0)

        # apply uncorrelated color noise
        d = {x.id: x.color_jitter for x in self.biomes}
        u, inv = np.unique(biome_ids, return_inverse=True)
        jitters = np.array([d[x] for x in u])[inv].reshape(biome_ids.shape)
        jitters = jitters.reshape(*jitters.shape, 1).repeat(3, axis=-1)
        return np.clip(colors + rand.normal(0.0, jitters, colors.shape), 0.0, 1.0)  # type: ignore

    def copy(self) -> "BiomeMap":
        return BiomeMap(
            map=self.map.copy(),
            biome_id=self.biome_id.copy(),
            border_distances=self.border_distances.copy(),
            biomes=self.biomes,
            config=self.config,
        )

    def _resolve_desired_point_count(
        self,
        density: float,
        biome_weights: np.ndarray,
    ) -> int:
        square_meters_per_cell = (1.0 / self.map.cells_per_meter) ** 2
        square_meters = biome_weights.sum() * square_meters_per_cell
        desired_point_count = round(square_meters * density)
        return desired_point_count

    def get_random_points(
        self,
        rand: np.random.Generator,
        desired_point_count: int,
        biome_weights: np.ndarray,
        is_debug_graph_printing_enabled: bool = False,
    ) -> np.ndarray:
        # because of the way we randomize below, we just ignore the final colums
        biome_weights = biome_weights.copy()
        biome_weights[:, -1] = 0.0
        biome_weights[-1, :] = 0.0

        # TODO: how we efficiently grab a bunch of random points really depends on
        #  how many points we want, how big the region is, etc. If we're trying to get
        #  multiple points per grid cell, we're likely better off with a different approach
        cell_locations = np.argwhere(biome_weights)
        selected_cells = rand.choice(
            cell_locations,
            p=biome_weights[np.nonzero(biome_weights)] / biome_weights.sum(),
            size=desired_point_count,
            shuffle=False,
        )

        # TODO: actually, just set ones wherever there is anything
        if is_debug_graph_printing_enabled:
            item_placements = np.zeros_like(biome_weights)
            for location in selected_cells:
                item_placements[tuple(location)] += 1
            plot_value_grid(item_placements, "biome item placements")

        points_2d = np.zeros_like(selected_cells, dtype=float)
        cell_size = 1.0 / self.map.cells_per_meter
        points_2d[:, 1] = (
            self.map.region.x.min_ge
            + selected_cells[:, 0] * cell_size
            + rand.uniform(0, cell_size, (desired_point_count))
        )
        points_2d[:, 0] = (
            self.map.region.z.min_ge
            + selected_cells[:, 1] * cell_size
            + rand.uniform(0, cell_size, (desired_point_count))
        )
        heights = self.map.get_heights(points_2d)
        return np.stack([points_2d[:, 0], heights, points_2d[:, 1]], axis=1)

    def get_random_points_in_biome(
        self,
        rand: np.random.Generator,
        config: SceneryConfig,
        placement_noise: Optional[MapFloatNP] = None,
        is_debug_graph_printing_enabled: bool = False,
        custom_weights: Optional[MapFloatNP] = None,
    ) -> List[Scenery]:
        if not self.config.is_scenery_added:
            return []
        if custom_weights is None:
            weights = np.ones_like(self.biome_id, dtype=np.float32)
            weights[self.biome_id != config.biome_id] = 0.0
            weights *= self.border_distances
            if config.border_mode == SceneryBorderMode.HARD:
                weights[weights < config.border_distance] = 0.0
                weights[weights > 0.0] = 1.0
            elif config.border_mode == SceneryBorderMode.LINEAR:
                weights = np.clip(weights / config.border_distance, 0, 1)
            elif config.border_mode == SceneryBorderMode.SQUARED:
                weights = np.clip(weights / config.border_distance, 0, 1) ** 2
            else:
                raise SwitchError("Mode not implemented yet: " + str(config.border_mode))
        else:
            weights = custom_weights

        if placement_noise is not None:
            weights *= placement_noise

        if is_debug_graph_printing_enabled:
            plot_value_grid(weights, "biome item placement weights")

        desired_point_count = self._resolve_desired_point_count(config.density, weights)
        if desired_point_count == 0:
            return []

        points = self.get_random_points(rand, desired_point_count, weights, is_debug_graph_printing_enabled)

        scales = np.ones((desired_point_count, 3), dtype=np.float32)
        correlated_scale_noise = rand.uniform(
            -config.correlated_scale_range, config.correlated_scale_range, size=(desired_point_count, 1)
        ).repeat(3, axis=1)
        uncorrelated_scale_noise = rand.uniform(
            -config.skewed_scale_range, config.skewed_scale_range, (desired_point_count, 3)
        )
        scales = scales + correlated_scale_noise + uncorrelated_scale_noise

        # bump things down into the ground, must consider scale as well
        flora_config, resource_name = get_flora_config_by_file(self.config.flora_config, config.resource_file)
        if flora_config is not None:
            points[:, 1] += flora_config.height_offset * scales[:, 1] * flora_config.default_scale

        results = []

        if config.is_oriented_to_surface:
            normals = self.map.get_normals(points)
            for point, scale, normal in zip(points, scales, normals):
                rotation = rotation_matrix_from_vectors(normal).flatten()
                results.append(
                    Scenery(resource_file=config.resource_file, position=point, rotation=rotation, scale=scale)
                )
        else:
            rotation_indices = rand.integers(0, 360, (desired_point_count))
            for point, scale, rotation_index in zip(points, scales, rotation_indices):
                rotation = _cached_rotations[rotation_index]
                results.append(
                    Scenery(resource_file=config.resource_file, position=point, rotation=rotation, scale=scale)
                )

        return results

    def create_extra_height_points(
        self,
        rand: np.random.Generator,
        is_climbable: MapBoolNP,
        is_detail_important: MapBoolNP,
    ) -> Point2DListNP:
        meters_per_cell = 1.0 / self.map.cells_per_meter
        square_meters_per_cell = meters_per_cell**2
        # points_per_cell = round(point_density_in_points_per_square_meter * square_meters_per_cell)
        # if points_per_cell == 0:
        #     points_per_cell = 1

        is_detail_important = morphology.dilation(is_detail_important, morphology.disk(2))
        # plot_value_grid(is_detail_important)

        points_2d = np.stack([self.map.X, self.map.Y], axis=2)
        base_points = points_2d[is_detail_important].reshape((-1, 2))
        noise_scale = meters_per_cell * 0.2
        half_point = np.array([meters_per_cell / 2.0, meters_per_cell / 2.0])
        noised_important_points = np.concatenate(
            [
                base_points + rand.uniform(-noise_scale, noise_scale, base_points.shape),
                base_points + rand.uniform(-noise_scale, noise_scale, base_points.shape) + half_point,
            ],
            axis=0,
        )

        # create extra points where it is very steep as well
        squared_slope = self.map.get_squared_slope()
        # plot_value_grid(squared_slope >= 16)
        # plot_value_grid(squared_slope >= 9)
        # plot_value_grid(squared_slope >= 4)
        # plot_value_grid(squared_slope >= 1)
        # this actually works out pretty reasonably, perhaps could use a bit more tuning
        steep_points = squared_slope >= 4

        # we could select more or fewer points here, but just selecting them all seems to work out for nwo
        unclimbable_mask = np.logical_or(steep_points, np.logical_not(is_climbable))
        unclimbable_points = points_2d[np.logical_and(unclimbable_mask, np.logical_not(is_detail_important))].reshape(
            (-1, 2)
        )
        rand.shuffle(unclimbable_points, axis=0)
        unclimbable_point_count = round(len(unclimbable_points) * 0.2)
        selected_unclimbable_points = unclimbable_points[:unclimbable_point_count]
        noise_scale = meters_per_cell * 0.3
        noised_unclimbable_points = selected_unclimbable_points + rand.uniform(
            -noise_scale, noise_scale, selected_unclimbable_points.shape
        )

        result = np.concatenate([noised_important_points, noised_unclimbable_points], axis=0)

        result[:, 0] = np.clip(result[:, 0], self.map.region.x.min_ge, self.map.region.x.max_lt)
        result[:, 1] = np.clip(result[:, 1], self.map.region.z.min_ge, self.map.region.z.max_lt)
        return cast(Point2DListNP, result)

    def plot(self) -> None:
        self.map.plot()


_IDENTITY = np.array([1, 0, 0, 0, 1, 0, 0, 0, 1])
_IDENTITY_MATRIX = np.eye(3)


# TODO: could be vectorized
def rotation_matrix_from_vectors(b: np.ndarray):
    a = UP_VECTOR
    v = np.cross(a, b)
    c = np.dot(a, b)
    s = np.linalg.norm(v)
    kmat = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
    rotation_matrix = _IDENTITY_MATRIX + kmat + kmat.dot(kmat) * ((1 - c) / (s**2))
    return rotation_matrix


_cached_rotations = [Rotation.from_euler("y", i, degrees=True).as_matrix().flatten() for i in range(360)]


def make_fast_biome(map: HeightMap, natural_biome_config: BiomeConfig):
    border_distances = np.ones_like(map.Z)
    biome_id = np.ones(map.Z.shape, dtype=np.int8)
    coast_biome_id = 1
    biomes = tuple(
        [
            Biome(coast_biome_id, "coastal", "#FFF18E", color_jitter=0.02),
            Biome(UNCLIMBABLE_BIOME_ID, "unclimbable", "#FF0000", color_jitter=0.02),
        ]
    )

    config = BiomeConfig(
        low_slope_threshold=1.0,
        high_slope_threshold=4.0,
        min_elevation=0.0,
        max_elevation=0.2,
        min_dryness=0.3,
        max_dryness=1.0,
        water_cutoff=0.9,
        noise_scale=0.01,
        noise_magnitude=1.0,
        is_color_per_vertex=False,
        is_normal_per_vertex=False,
        color_map={x.id: hex_to_rgb(x.color) for x in biomes},
        biome_matrix=_raw_biome_id_mapping,
        biomes=biomes,
        flora_config=natural_biome_config.flora_config,
        godot_sky_config=natural_biome_config.godot_sky_config,
        godot_sun_config=natural_biome_config.godot_sun_config,
        godot_env_config=_disable_env_settings(natural_biome_config.godot_env_config),
        coastal_id=coast_biome_id,
        rock_biome_ids=tuple(),
        color_sampling_neighbor_count=0,
        is_scenery_added=natural_biome_config.is_scenery_added,
    )
    return BiomeMap(map, biomes, biome_id, border_distances, config)


def _disable_env_settings(env_settings: Dict[str, Any]) -> Dict[str, Any]:
    return {
        k: (v if (not isinstance(v, bool) or not k.endswith("_enabled")) else False) for k, v in env_settings.items()
    }


def make_biome(
    map: HeightMap,
    config: BiomeConfig,
    rand: np.random.Generator,
    special_biomes: Optional[SpecialBiomes] = None,
    is_debug_graph_printing_enabled: bool = False,
) -> BiomeMap:
    water_distance = map.get_water_distance(
        rand, config.is_fresh_water_included_in_moisture, max_points=config.max_kd_points
    )
    if water_distance is None:
        moisture = np.zeros_like(map.Z)
        water_distance = np.full_like(map.Z, np.inf)
    else:
        moisture = 1.0 - (water_distance / water_distance.max(initial=0.001))

    if is_debug_graph_printing_enabled:
        height_with_water = np.clip(map.Z, WATER_LINE, map.Z.max(initial=1.0))
        plot_value_grid(height_with_water, "height map")
        plot_value_grid(moisture, "moisture map")

    elevation = map.get_elevation()
    min_elevation = WATER_LINE
    normalized_elevation = (elevation - min_elevation) / (elevation.max(initial=0.001) - min_elevation)
    elevation_range = config.max_elevation - config.min_elevation
    normalized_elevation = normalized_elevation * elevation_range + config.min_elevation

    dryness = 1.0 - moisture.copy()
    noise_data = perlin(elevation.shape, config.noise_scale, rand)
    dryness = np.clip(dryness * (1 + noise_data * config.noise_magnitude), 0, 1)
    _clamp_water(dryness, moisture, config)

    squared_slope = map.get_squared_slope()

    biome_id = _map_to_biome_id(
        normalized_elevation,
        dryness,
        moisture,
        elevation,
        water_distance,
        squared_slope,
        config,
        special_biomes=special_biomes,
    )

    border_distances = _get_border_distances(rand, biome_id, map.cells_per_meter, config.max_kd_points)

    biome_map = BiomeMap(map, config.biomes, biome_id, border_distances, config)

    if is_debug_graph_printing_enabled:
        raw_dryness = 1.0 - moisture.copy()
        _clamp_water(raw_dryness, moisture, config)
        biome_id_without_noise = _map_to_biome_id(
            normalized_elevation,
            raw_dryness,
            moisture,
            elevation,
            water_distance,
            squared_slope,
            config,
            special_biomes=special_biomes,
        )
        plot_biome_grid(biome_id_without_noise, config, "biome map (without noise)")

        fig, ax = plt.subplots()
        fig.set_size_inches(8, 8)
        ax.set_title("noise texture")
        plt.imshow(noise_data, origin="lower")
        plt.show()
        fig.clf()

        plot_biome_grid(biome_map.biome_id, config, "biome map (with noise)")

        plot_value_grid(border_distances, "distance from biome edge")

        plot_biome_grid(config.biome_matrix[0], config, "low steepness biome mapping by dryness (X) and elevation (Y)")
        plot_biome_grid(
            config.biome_matrix[1], config, "medium steepness biome mapping by dryness (X) and elevation (Y)"
        )
        plot_biome_grid(
            config.biome_matrix[2], config, "high steepness biome mapping by dryness (X) and elevation (Y)"
        )
        print_biome_legend(config)

    return biome_map


def _clamp_water(dryness: MapFloatNP, moisture: MapFloatNP, config: BiomeConfig) -> None:
    water_cutoff = config.water_cutoff
    dryness[moisture < water_cutoff] = np.clip(
        dryness[moisture < water_cutoff], config.min_dryness, config.max_dryness
    )
    dryness[moisture >= water_cutoff] = 0.0


def _get_border_distances(
    rand: np.random.Generator, biome_id: MapIntNP, cells_per_meter: float, max_points: Optional[int]
) -> MapFloatNP:
    border, border_points = get_border_points(biome_id, rand)

    distances = np.ones_like(biome_id, dtype=np.float32)
    if len(border_points) == 0:
        return distances

    distances[border] = 0.0
    tree = create_kd_tree(border_points, "borders", max_points=max_points, rand=rand)
    distances[np.nonzero(distances)] = tree.query(np.argwhere(distances), k=1)[0]

    return distances * (1.0 / cells_per_meter)


def _map_to_biome_id(
    normalized_elevation: MapFloatNP,
    dryness: MapFloatNP,
    moisture: MapFloatNP,
    elevation: MapFloatNP,
    water_distance: MapFloatNP,
    squared_slope: MapFloatNP,
    config: BiomeConfig,
    special_biomes: Optional[SpecialBiomes] = None,
) -> MapIntNP:
    elevation_bins = config.biome_matrix[0].shape[0]
    elevation_indices = np.clip((normalized_elevation * elevation_bins - 0.5).astype(int), 0, elevation_bins - 1)
    dryness_bins = config.biome_matrix[0].shape[1]
    dryness_indices = np.clip((dryness * dryness_bins - 0.5).astype(int), 0, dryness_bins - 1)

    low_mask = squared_slope < config.low_slope_threshold
    med_mask = np.logical_and(config.low_slope_threshold <= squared_slope, squared_slope < config.high_slope_threshold)
    high_mask = config.high_slope_threshold <= squared_slope
    biome_id = np.zeros_like(dryness_indices)
    for i, mask in enumerate([low_mask, med_mask, high_mask]):
        biome_id[mask] = config.biome_matrix[i][elevation_indices, dryness_indices][mask]

    if special_biomes:
        biome_id[special_biomes.swamp_mask] = SWAMP

    beach_mask = np.full_like(biome_id, False)
    if special_biomes:
        beach_mask = special_biomes.beach_mask
        biome_id[beach_mask] = 3

    # the dark sand region is everything within a pretty small distance of the beach line
    dark_beach_mask = np.logical_and(
        beach_mask,
        np.logical_and(water_distance < config.dark_shore_distance, elevation < WATER_LINE + config.dark_shore_height),
    )
    biome_id[dark_beach_mask] = DARK_SHORE

    # reset anything that is completely wet to be water
    biome_id[moisture == 1.0] = 2

    # set fresh water and swamp masks
    if special_biomes:
        biome_id[special_biomes.fresh_water_mask] = FRESH_WATER

    # expand and contract rocks so that they are not scattered
    rock_id = only(config.rock_biome_ids)
    base_rock_ids = biome_id == rock_id
    expander = morphology.disk(4)
    true_rock_ids = np.logical_and(
        base_rock_ids,
        morphology.dilation(
            np.logical_and(
                base_rock_ids,
                np.logical_not(morphology.dilation(np.logical_not(base_rock_ids), expander)),
            ),
            expander,
        ),
    )
    removed_rock_ids = np.logical_and(base_rock_ids, np.logical_not(true_rock_ids))
    biome_id[removed_rock_ids] = config.biome_matrix[0][elevation_indices, dryness_indices][removed_rock_ids]

    # and reset things that are just always cliffs
    cliffs = squared_slope > config.high_slope_threshold
    # cliffs = squared_slope > config.force_cliff_square_slope
    # cliffs = morphology.dilation(cliffs, morphology.disk(1))
    biome_id[cliffs] = rock_id

    # plot_value_grid(biome_id == rock_id, "ROCK IDS")

    return cast(MapIntNP, biome_id)


def plot_biome_grid(data: np.ndarray, config: BiomeConfig, title: str = "") -> None:
    data = np.array(data)
    fig, ax = plt.subplots()
    fig.set_size_inches(8, 8)
    if title:
        ax.set_title(title)

    image_data = []
    for row in data:
        new_row = []
        for value in row:
            new_row.append(config.color_map[value])
        image_data.append(new_row)

    image_data_np = np.array(image_data)
    plt.imshow(image_data_np, origin="lower")
    plt.show()


def print_biome_legend(config: BiomeConfig) -> None:
    display(HTML(f"<h1>Biomes:</h1>"))

    for biome in config.biomes:
        display(HTML(f'<h2 style="background: {biome.color}">{biome.name}</h2>'))



from typing import Optional
from typing import Tuple

import gym
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from avalon.common.errors import SwitchError


def init(module, weight_init, bias_init, gain=1):
    """Helper to initialize a layer weight and bias."""
    weight_init(module.weight.data, gain=gain)
    bias_init(module.bias.data)
    return module


def mlp_init(module, gain=np.sqrt(2), bias=0.0):
    """Helper to initialize a layer weight and bias."""
    nn.init.orthogonal_(module.weight.data, gain=gain)
    nn.init.constant_(module.bias.data, bias)
    return module


class Flatten(nn.Module):
    def forward(self, x):
        return x.reshape(x.size(0), -1)


class CNNBase(nn.Module):
    def __init__(self, observation_space, action_space, hidden_size=512) -> None:
        """Initializer.
        num_channels: the number of channels in the input images (eg 3
            for RGB images, or 12 for a stack of 4 RGB images).
        num_outputs: the dimension of the output distribution.
        dist: the output distribution (eg Discrete or Normal).
        hidden_size: the size of the final actor+critic linear layers

        TODO: this needs updated to the new action types.
        """
        super().__init__()

        # TODO: make this more general to allow non-scalar spaces
        assert len(observation_space["wrapped"].shape) == 3
        num_channels = observation_space["wrapped"].shape[0]
        self.num_outputs = len(action_space.spaces)

        init_ = lambda m: init(
            m, nn.init.orthogonal_, lambda x: nn.init.constant_(x, 0), nn.init.calculate_gain("relu")
        )

        self.main = nn.Sequential(
            init_(nn.Conv2d(num_channels, 32, kernel_size=8, stride=4)),
            nn.ReLU(),
            init_(nn.Conv2d(32, 64, kernel_size=4, stride=2)),
            nn.ReLU(),
            init_(nn.Conv2d(64, 32, kernel_size=3, stride=1)),
            nn.ReLU(),
            Flatten(),
            # init_(nn.Linear(32 * 7 * 7, hidden_size)),
            # nn.ReLU(),
            mlp_init(nn.Linear(32 * 7 * 7, hidden_size)),
            nn.Tanh(),
        )

        init_ = lambda m: init(m, nn.init.orthogonal_, lambda x: nn.init.constant_(x, 0))

        self.critic_linear = mlp_init(nn.Linear(hidden_size, 1), gain=1.0)
        self.action_head = DictActionHead(action_space)
        self.actor_linear = mlp_init(nn.Linear(hidden_size, self.action_head.num_inputs), gain=0.01)

    def forward(self, obs):
        """x should have shape (batch_size, num_channels, 84, 84)."""
        x = obs["wrapped"]
        batch_size = x.shape[0]
        x = self.main(x)
        value = self.critic_linear(x).reshape((batch_size,))
        action_logits = self.actor_linear(x)
        assert value.shape == (batch_size,)
        assert action_logits.shape == (batch_size, self.action_head.num_inputs)
        return value, self.action_head(action_logits)


class MLPBase(nn.Module):
    def __init__(self, observation_space, action_space, hidden_size=64) -> None:
        super().__init__()

        # TODO: make this more general to allow non-scalar spaces
        assert len(observation_space["wrapped"].shape) == 1
        num_inputs = observation_space["wrapped"].shape[0]

        self.action_head = DictActionHead(action_space)

        self.actor = nn.Sequential(
            mlp_init(nn.Linear(num_inputs, hidden_size)),
            nn.Tanh(),
            mlp_init(nn.Linear(hidden_size, hidden_size)),
            nn.Tanh(),
            mlp_init(nn.Linear(hidden_size, self.action_head.num_inputs), gain=0.01),
        )

        self.critic = nn.Sequential(
            mlp_init(nn.Linear(num_inputs, hidden_size)),
            nn.Tanh(),
            mlp_init(nn.Linear(hidden_size, hidden_size)),
            nn.Tanh(),
            mlp_init(nn.Linear(hidden_size, 1), gain=1.0),
        )

    def forward(self, obs):
        x = obs["wrapped"]
        batch_size = x.shape[0]
        value = self.critic(x).reshape((batch_size,))
        action_logits = self.actor(x)
        assert value.shape == (batch_size,)
        assert action_logits.shape == (batch_size, self.action_head.num_inputs)
        return value, self.action_head(action_logits)


# TODO: this needs fixed to take a 2-d policy logit
class DiscreteHead(nn.Module):
    """A module that builds a Categorical distribution from logits."""

    def __init__(self, num_outputs, **kwargs) -> None:
        super().__init__()
        self.num_inputs = num_outputs
        self.num_outputs = 1

    def forward(self, x):
        # x should have shape (batch_size, num_actions, num_categories)
        assert x.shape[-1] == self.num_inputs
        probs = F.softmax(x, dim=-1)
        dist = torch.distributions.Categorical(probs=probs)
        # Categorical already has only one output, no need for independent!
        # dist = torch.distributions.Independent(dist, reinterpreted_batch_ndims=num_batch_dims)
        return dist


class BernoulliHead(nn.Module):
    def __init__(self, n_outputs, bias: Optional[Tuple[float, ...]] = None, **kwargs) -> None:
        super().__init__()
        self.num_inputs = n_outputs
        self.num_outputs = n_outputs
        if bias is None:
            self.bias = tuple(0 for _ in range(n_outputs))
        else:
            assert len(bias) == n_outputs
            self.bias = bias

    def forward(self, x):
        # x should have shape (batch_size, 1)
        assert x.shape[-1] == self.num_inputs
        x = torch.clamp(x, -6, 6)
        bias = torch.tensor(self.bias).view([1] * (len(x.shape) - 1) + [self.num_inputs]).to(x.device)
        dist = torch.distributions.Bernoulli(logits=x + bias)
        dist = torch.distributions.Independent(dist, reinterpreted_batch_ndims=1)
        return dist


class NormalHead(nn.Module):
    """A module that builds a Diagonal Gaussian distribution from means.

    Standard deviations are learned parameters in this module.
    """

    def __init__(self, action_space: gym.spaces.Box, **kwargs) -> None:
        super().__init__()
        assert isinstance(action_space, gym.spaces.Box)
        assert len(action_space.shape) == 1
        self.num_outputs = action_space.shape[0]
        self.num_inputs = self.num_outputs * 2
        # initial variance is e^0 = 1
        # self.log_std = nn.Parameter(torch.zeros(self.num_outputs))
        self.register_buffer("center", torch.tensor((action_space.high + action_space.low) / 2))
        # TODO: should we put the scale into the initial value of log_std instead?
        # Would effect the effective LR of that parameter.
        self.register_buffer("scale", torch.tensor((action_space.high - action_space.low) / 2))

    def forward(self, x):
        # x should have shape (batch_size, action_dim)
        assert x.shape[-1] == self.num_inputs
        means, log_stds = torch.chunk(x, 2, dim=-1)
        log_stds = torch.clamp(log_stds, -6, 6)
        dist = torch.distributions.Normal(loc=means, scale=log_stds.exp())
        dist = torch.distributions.Independent(dist, reinterpreted_batch_ndims=1)
        return dist


class DictActionHead(torch.nn.Module):
    def __init__(self, action_space) -> None:
        super().__init__()
        self.action_space = action_space

        # Build action heads
        self.num_inputs = 0
        action_heads = {}
        for k, space in self.action_space.spaces.items():
            # TODO: major perf gains from not forcing boxes to have scalar shape
            if isinstance(space, gym.spaces.Box):
                action_heads[k] = NormalHead(space)
            elif isinstance(space, gym.spaces.Discrete):
                action_heads[k] = DiscreteHead(space.n)
            elif isinstance(space, gym.spaces.MultiBinary):
                action_heads[k] = BernoulliHead(space.n)
            else:
                assert False

            self.num_inputs += action_heads[k].num_inputs
        self.action_heads = torch.nn.ModuleDict(action_heads)

    def forward(self, action_logits):
        batch_size = action_logits.shape[:-1]
        assert action_logits.shape[-1] == self.num_inputs

        dists = {}
        # NOTE: this depends on the fact that dicts are ordered in recent CPython.
        i = 0
        for k, head in self.action_heads.items():
            # we want the action logit to have shape (batch_size, num_inputs)
            logits = action_logits[..., i : i + head.num_inputs]
            assert logits.shape[-1] == head.num_inputs
            dists[k] = head(logits)
            i += head.num_outputs
        return DictActionDist(dists, self.action_heads)


class DictActionDist(torch.distributions.Distribution):
    def __init__(self, dists, action_heads) -> None:
        super().__init__(validate_args=False)
        self.dists = dists
        self.action_heads = action_heads

    def sample(self, device="cpu"):
        """Returns an action as a dict of batched tensors."""
        # rsample is not used - gradients propagate thru log_prob, not sample
        out = {}
        for k, v in self.dists.items():
            # should have shape (batch_size, num_outputs)
            sample = v.sample().to(device=device)
            # assert len(sample.shape) == 2
            # assert sample.dtype == torch.float32
            # assert sample.shape[-1] == self.action_heads[k].num_outputs
            out[k] = sample
        return out

    def log_prob(self, actions):
        """Compute the log prob of the given action under the given dist (batchwise).

        Log prob is a scalar (per batch element.)"""
        # actions is a dict of tensors of shape (batch_size, num_outputs)
        log_probs = []
        for k, dist in self.dists.items():
            # TODO: this used to call .float() . . . why?
            log_prob = dist.log_prob(actions[k])
            log_probs.append(log_prob)

        return torch.stack(log_probs, dim=-1).sum(dim=-1)

    def entropy(self):
        entropies = []
        for k, dist in self.dists.items():
            entropy = dist.entropy()
            # The entropy is over the entire action space, so it reduces to a single scalar per batch element
            assert len(entropy.shape) == 1
            entropies.append(entropy)
        return torch.stack(entropies, dim=1).sum(dim=1)

    def sample_mode(self, device="cpu"):
        out = {}
        for k, v in self.dists.items():
            if isinstance(v, torch.distributions.Independent):
                v = v.base_dist
            if isinstance(v, torch.distributions.Normal):
                sample = v.mean
            elif isinstance(v, torch.distributions.Categorical):
                _value, sample = torch.max(v.logits, dim=-1)
            elif isinstance(v, torch.distributions.Bernoulli):
                sample = torch.where(
                    v.logits > 0,
                    torch.ones_like(v.logits, dtype=torch.int),
                    torch.zeros_like(v.logits, dtype=torch.int),
                )
            else:
                raise SwitchError(f"Unknown distribution for mode, {v.__class__}")
            out[k] = sample.to(device=device)
        return out



from pathlib import Path
from typing import Final
from typing import List
from typing import Type

import attr
import numpy as np
from scipy import stats

from avalon.datagen.errors import WorldTooSmall
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.task import TaskConfig
from avalon.datagen.world_creation.constants import MIN_FALL_DISTANCE_TO_CAUSE_DAMAGE
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.entities.animals import Crow
from avalon.datagen.world_creation.entities.animals import Deer
from avalon.datagen.world_creation.entities.animals import Mouse
from avalon.datagen.world_creation.entities.animals import Pigeon
from avalon.datagen.world_creation.entities.animals import Prey
from avalon.datagen.world_creation.entities.animals import Rabbit
from avalon.datagen.world_creation.entities.animals import Squirrel
from avalon.datagen.world_creation.entities.animals import Turtle
from avalon.datagen.world_creation.entities.tools.weapons import Rock
from avalon.datagen.world_creation.utils import to_2d_point
from avalon.datagen.world_creation.worlds.creation import create_world_from_constraint
from avalon.datagen.world_creation.worlds.difficulty import normal_distrib_range
from avalon.datagen.world_creation.worlds.difficulty import scale_with_difficulty
from avalon.datagen.world_creation.worlds.difficulty import select_boolean_difficulty
from avalon.datagen.world_creation.worlds.difficulty import select_categorical_difficulty
from avalon.datagen.world_creation.worlds.export import export_world
from avalon.datagen.world_creation.worlds.obstacles.configure import make_ring
from avalon.datagen.world_creation.worlds.obstacles.height_solution import HeightSolution
from avalon.datagen.world_creation.worlds.utils import add_offsets

# a little wiggle room so that the pit and weapons and everything fit properly
_SAFETY_RADIUS = 1.0
# must have at least this much space
_MIN_OBSTACLE_SIZE = 0.5


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class ThrowTaskConfig(TaskConfig):
    # how far away the goal should be for difficulty 0.0 and 1.0 respectively
    # this implicitly controls the size of the pit, along with the next set of parameters
    goal_dist_easy: float = 5.0
    goal_dist_hard: float = 20.0
    # in pit mode, controls how much of the goal distance ends up being used in the pit vs outside the pit
    # by using less of the distance in the pit, the pit is smaller, which is easier, since the target cannot move
    # around as much
    pit_distance_fraction_easy: float = 0.1
    pit_distance_fraction_hard: float = 0.9
    pit_distance_std_dev: float = 0.5
    # how deep to make the pit
    pit_depth_easy: float = 3.0
    pit_depth_hard: float = MIN_FALL_DISTANCE_TO_CAUSE_DAMAGE
    # how many rocks are spawned at difficulty 0.0 and 1.0 respectively
    rock_count_easy: int = 6
    rock_count_hard: int = 2
    rock_count_std_dev: float = 0.25
    # the edge of the pit will span this much distance horizontally. Controls how steep it is
    incline_distance = 1.0
    # controls how far away from the edge of the pit you start
    brink_distance = 0.5


ALL_PREY_CLASSES_FOR_THROW: Final[List[Type[Prey]]] = [
    Turtle,
    Mouse,
    Rabbit,
    Squirrel,
    Deer,
    Pigeon,
    Crow,
]


def generate_throw_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: ThrowTaskConfig = ThrowTaskConfig(),
) -> None:

    difficulty, food_class, locations, spawn_location, world = throw_obstacle_maker(
        rand, difficulty, export_config, task_config=task_config
    )
    world, locations = world.end_height_obstacles(locations, is_accessible_from_water=False)
    world = world.add_spawn_and_food(rand, difficulty, spawn_location, locations.goal, food_class=food_class)
    export_world(output_path, rand, world)


def throw_obstacle_maker(
    rand: np.random.Generator,
    difficulty: float,
    export_config: ExportConfig,
    is_for_carry: bool = False,
    task_config: ThrowTaskConfig = ThrowTaskConfig(),
):
    food_class, output_difficulty = select_categorical_difficulty(ALL_PREY_CLASSES_FOR_THROW, difficulty, rand)
    offset = food_class(position=np.array([0.0, 0.0, 0.0])).get_offset()
    predator_type = None

    if is_for_carry:
        is_pit = True
    else:
        is_pit, output_difficulty = select_boolean_difficulty(difficulty, rand, final_prob=0.1)
    # this isn't exactly distance... because we reset the spawn position to be at the ledge below
    # so we will actually end up closer
    desired_goal_dist = scale_with_difficulty(difficulty, task_config.goal_dist_easy, task_config.goal_dist_hard)

    world, locations = create_world_from_constraint(
        stats.norm(desired_goal_dist, desired_goal_dist / 10), rand, difficulty, export_config, None, offset
    )

    actual_goal_distance = locations.get_2d_spawn_goal_distance()
    max_possible_outer_radius = (actual_goal_distance - (_SAFETY_RADIUS * 2 + task_config.incline_distance)) * 0.99

    if max_possible_outer_radius < _MIN_OBSTACLE_SIZE:
        raise WorldTooSmall(AvalonTask.THROW, _MIN_OBSTACLE_SIZE, locations.get_2d_spawn_goal_distance())

    outer_radius = 0.0
    if is_pit:
        # can be increased up to max_possible_outer_radius
        # makes the task harder because then the pit is bigger
        extra_pit_size = normal_distrib_range(
            max_possible_outer_radius * task_config.pit_distance_fraction_easy,
            max_possible_outer_radius * task_config.pit_distance_fraction_hard,
            task_config.pit_distance_std_dev,
            rand,
            difficulty,
        )
        outer_radius = max_possible_outer_radius - extra_pit_size

        # logger.debug(f"{extra_pit_size} out of {max_possible_outer_radius}")

        # disables extra large circles so that we can make very small pits and outcroppings
        max_additional_radius_multiple = 1.0
    else:
        # larger platforms are probably slightly easier, but doesn't make a huge difference
        max_additional_radius_multiple = 1.0

    rock_count = round(
        normal_distrib_range(
            task_config.rock_count_easy + 0.49,
            task_config.rock_count_hard - 0.49,
            task_config.rock_count_std_dev,
            rand,
            difficulty,
        )
    )
    rocks = [Rock(position=np.array([-0.5 * i, 0.0, 0.0])) for i in range(rock_count)]
    depth = scale_with_difficulty(difficulty, task_config.pit_depth_easy, task_config.pit_depth_hard)
    ring_config = make_ring(
        rand,
        difficulty,
        world,
        locations,
        gap_distance=0.0,
        height=-depth,
        traversal_width=10.0 - 9.0 * difficulty * rand.uniform(),
        inner_traversal_length=task_config.incline_distance,
        is_single_obstacle=True,
        inner_solution=HeightSolution(
            inside_items=tuple(add_offsets(rocks)),
            inside_item_randomization_distance=2.0,
            inside_item_radius=rocks[0].get_offset(),
            solution_point_brink_distance=task_config.brink_distance,
        ),
        probability_of_centering_on_spawn=0.0 if is_pit else 1.0,
        outer_traversal_length=outer_radius,
        max_additional_radius_multiple=max_additional_radius_multiple,
        constraint=None,
    )
    world = world.add_height_obstacle(rand, ring_config, locations.island)
    # reset the spawn location to the edge of the pit so we can see the food
    average_rock_position = np.array([x.position for x in world.items if isinstance(x, Rock)]).mean(axis=0)
    # height will be reset below
    spawn_location = average_rock_position
    if predator_type is not None:
        world = world.add_random_predator_near_point(
            rand,
            predator_type,
            to_2d_point(locations.goal),
            task_config.incline_distance + _SAFETY_RADIUS,
            locations.island,
        )
    return difficulty, food_class, locations, spawn_location, world



import sys

from fire import helptext


def make_fire_more_friendly(line_length: int = 200) -> None:
    # make fire stop truncating so many things
    helptext.LINE_LENGTH = line_length

    # this makes fire do what I would expect if you pass --help into the base command
    if len(sys.argv) == 2 and sys.argv[-1] in ("--help", "-h"):
        sys.argv.pop()



from pathlib import Path
from typing import Optional

import attr
import numpy as np

from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.worlds.compositional import CompositeTaskConfig
from avalon.datagen.world_creation.worlds.compositional import ForcedComposition
from avalon.datagen.world_creation.worlds.compositional import create_compositional_task
from avalon.datagen.world_creation.worlds.export import export_world


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class NavigateTaskConfig(CompositeTaskConfig):
    task: AvalonTask = AvalonTask.NAVIGATE


def generate_navigate_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: NavigateTaskConfig = NavigateTaskConfig(),
    _FORCED: Optional[ForcedComposition] = None,
) -> None:
    world, locations = create_compositional_task(rand, difficulty, task_config, export_config, _FORCED=_FORCED)
    export_world(output_path, rand, world)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import multiprocessing as mp
import os.path
import shutil
import tarfile
from pathlib import Path
from uuid import uuid4

import numpy as np
import wandb

from avalon.agent.evaluation import EVAL_TEMP_PATH
from avalon.agent.godot.godot_gym import CURRICULUM_BASE_PATH
from avalon.agent.torchbeast import polybeast_env
from avalon.agent.torchbeast import polybeast_learner
from avalon.agent.torchbeast.avalon_helpers import TORCHBEAST_ENV_LOGS_PATH
from avalon.agent.torchbeast.avalon_helpers import _destroy_process_tree
from avalon.common.log_utils import configure_parent_logging
from avalon.common.log_utils import logger
from avalon.contrib.s3_utils import SimpleS3Client
from avalon.contrib.utils import FILESYSTEM_ROOT


def run_env(flags, actor_id, exit_event) -> None:
    configure_parent_logging()
    np.random.seed()  # Get new random seed in forked process.
    polybeast_env.main(flags, exit_event)


def run_learner(flags):
    return polybeast_learner.main(flags)


def get_flags(check_argv_empty: bool = False):
    flags = argparse.Namespace()
    flags, argv = polybeast_learner.parser.parse_known_args(namespace=flags)
    flags, argv = polybeast_env.parser.parse_known_args(args=argv, namespace=flags)
    if check_argv_empty and argv:
        # Produce an error message.
        polybeast_learner.parser.print_usage()
        logger.info("")
        polybeast_env.parser.print_usage()
        logger.info("Unknown args:", " ".join(argv))
        return -1
    return flags


def restore_curriculum_files() -> None:
    api = wandb.Api()
    run_api = api.run(wandb.run.path)
    curriculum_filenames = [file.name for file in run_api.files() if file.name.startswith("curriculum")]
    os.makedirs(CURRICULUM_BASE_PATH, exist_ok=True)
    shutil.rmtree(CURRICULUM_BASE_PATH)
    for filename in curriculum_filenames:
        wandb.restore(filename, root=CURRICULUM_BASE_PATH.parent, replace=True)


def main(flags):
    if flags.fixed_world_key and flags.fixed_world_path:
        logger.info(f"Extracting {flags.fixed_world_key} to {flags.fixed_world_path}")
        s3_client = SimpleS3Client()
        os.makedirs(flags.fixed_world_path, exist_ok=True)
        os.makedirs(EVAL_TEMP_PATH, exist_ok=True)
        s3_client.download_to_file(flags.fixed_world_key, Path(EVAL_TEMP_PATH) / flags.fixed_world_key)
        with tarfile.open(Path(EVAL_TEMP_PATH) / flags.fixed_world_key, "r:gz") as f:
            f.extractall(flags.fixed_world_path)
    else:
        data_path = f"{FILESYSTEM_ROOT}/data/level_gen"
        os.makedirs(data_path, exist_ok=True)
        shutil.rmtree(data_path)

    os.makedirs(TORCHBEAST_ENV_LOGS_PATH, exist_ok=True)
    shutil.rmtree(TORCHBEAST_ENV_LOGS_PATH)

    if not flags.xpid:
        flags.xpid = str(uuid4())
    name = os.environ.get("EXPERIMENT_NAME", flags.env)

    if flags.mode == "train":
        wandb.init(
            config=flags,
            project=flags.project,
            name=name + f"__{flags.training_protocol}",
            id=flags.xpid,
            resume="allow",
        )
        restore_curriculum_files()
    else:
        wandb.init(
            config=flags, project=flags.project, name=name + f"__{flags.mode}", notes=f"{flags.project}/{flags.xpid}"
        )

    env_processes = []
    exit_event = mp.Event()
    for actor_id in range(1):
        p = mp.Process(target=run_env, args=(flags, actor_id, exit_event))
        p.start()
        env_processes.append(p)

    result = run_learner(flags)

    logger.info("Learner finished, killing processes")
    exit_event.set()

    for p in env_processes:
        _destroy_process_tree(p.pid)

    return result


if __name__ == "__main__":
    flags = get_flags(check_argv_empty=True)
    main(flags)



import math
from typing import List
from typing import Tuple
from typing import Union

import attr
import numpy as np
import openturns as ot
from numpy.random import Generator

from avalon.contrib.serialization import Serializable
from avalon.datagen.godot_base_types import FloatRange
from avalon.datagen.world_creation.utils import get_random_seed_for_line


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Region(Serializable):
    x: FloatRange
    z: FloatRange

    def contains_point_2d(self, point: Union[np.ndarray, Tuple[float, float]], epsilon: float = 0.01) -> bool:
        if point[0] >= self.x.min_ge - epsilon and point[0] < self.x.max_lt + epsilon:
            if point[1] >= self.z.min_ge - epsilon and point[1] < self.z.max_lt + epsilon:
                return True
        return False

    def contains_point_2d_exact(self, point: Union[np.ndarray, Tuple[float, float]]) -> bool:
        return (
            point[0] >= self.x.min_ge
            and point[0] < self.x.max_lt
            and point[1] >= self.z.min_ge
            and point[1] < self.z.max_lt
        )

    def overlaps_region(self, other: "Region") -> bool:
        for point in other.points:
            if self.contains_point_2d(point):
                return True
        if other.contains_point_2d((self.x.min_ge, self.z.min_ge)):
            return True
        return False

    def contains_point_3d(self, point: Union[np.ndarray, Tuple[float, float, float]], epsilon: float = 0.01) -> bool:
        if point[0] > self.x.min_ge - epsilon and point[0] < self.x.max_lt + epsilon:
            if point[2] > self.z.min_ge - epsilon and point[2] < self.z.max_lt + epsilon:
                return True
        return False

    def contains_region(self, other: "Region", epsilon: float = 0.01) -> bool:
        for point in other.points:
            if not self.contains_point_2d(point, epsilon=epsilon):
                return False
        return True

    @property
    def points(self) -> List[Tuple[float, float]]:
        return [
            (float(self.x.min_ge), float(self.z.min_ge)),
            (float(self.x.max_lt), float(self.z.min_ge)),
            (float(self.x.max_lt), float(self.z.max_lt)),
            (float(self.x.min_ge), float(self.z.max_lt)),
        ]

    def get_linspace(self, cells_per_meter: float) -> Tuple[np.ndarray, np.ndarray]:
        x = np.linspace(self.x.min_ge, self.x.max_lt, round(self.x.size * cells_per_meter) + 1)
        y = np.linspace(self.z.min_ge, self.z.max_lt, round(self.z.size * cells_per_meter) + 1)
        return x, y

    def get_random_generator(self, seed: int) -> Generator:
        seed_for_region = get_random_seed_for_line(
            seed, (self.x.min_ge, self.z.min_ge), (self.x.max_lt, self.z.max_lt)
        )
        return np.random.default_rng(seed_for_region)

    def epsilon_expand(self, x_min: bool, x_max: bool, z_min: bool, z_max: bool, epsilon: float = 0.01) -> "Region":
        x_min_epsilon = epsilon if x_min else 0
        x_max_epsilon = epsilon if x_max else 0
        z_min_epsilon = epsilon if z_min else 0
        z_max_epsilon = epsilon if z_max else 0
        return Region(
            FloatRange(self.x.min_ge - x_min_epsilon, self.x.max_lt + x_max_epsilon),
            FloatRange(self.z.min_ge - z_min_epsilon, self.z.max_lt + z_max_epsilon),
        )

    def get_randomish_points(self, count: int) -> np.ndarray:
        points = _create_low_discrepancy_random_points(count)
        # shift to 0 - 1 range
        points *= 0.5
        points += 0.5
        # scale to region size
        sizes = np.array([self.x.size, self.z.size])
        points *= sizes
        # add the min
        min_point = np.array([self.x.min_ge, self.z.min_ge])
        points += min_point
        return points


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class EdgedRegion(Region):
    edge_vertices: np.ndarray


def _create_low_discrepancy_random_points(min_point_count: int) -> np.ndarray:
    dim = 2
    distribution = ot.ComposedDistribution([ot.Uniform()] * dim)
    sequence = ot.SobolSequence(dim)
    samplesize = 2 ** (int(math.log2(min_point_count)) + 1)  # Sobol' sequences are in base 2
    experiment = ot.LowDiscrepancyExperiment(sequence, distribution, samplesize, False)
    sample = experiment.generate()
    return np.array(sample)



from enum import Enum
from typing import Dict
from typing import Tuple

import numpy as np

from avalon.common.errors import SwitchError
from avalon.datagen.world_creation.geometry import Axis
from avalon.datagen.world_creation.types import TileConvKernel

TILE_SIZE = 1
DEFAULT_STORY_HEIGHT = 5  # does not include floor thickness
DEFAULT_FLOOR_THICKNESS = 1
SLOPE_THICKNESS = 0.1
LADDER_THICKNESS = 0.1
CEILING_THICKNESS = 1
HIGH_POLY = "high_poly"
LOW_POLY = "low_poly"

# Rooms need at least 3 tiles to place a ladder and 1 tile on each side to have a wall (so 5 tiles for a building)
# We could have smaller one-story buildings, but it's nicer to keep things consistent and extensible
MIN_ROOM_SIZE = 3
MIN_BUILDING_SIZE = 5


class ExportMode(Enum):
    PRETTY = "PRETTY"
    DEBUG = "DEBUG"


EXPORT_MODE = ExportMode.PRETTY


class TileIdentity(Enum):
    FULL = 0
    ROOM = -1
    HALLWAY = -2
    LINK = -3
    LINK_BOTTOM_LANDING = -4
    LINK_TOP_LANDING = -5
    VOID = -6

    @property
    def pretty_name(self) -> str:
        return " ".join([section.lower() for section in self.name.split("_")])


class CornerType(Enum):
    NE = "NE"
    SE = "SE"
    SW = "SW"
    NW = "NW"

    @property
    def convolution_kernel(self) -> TileConvKernel:
        kernels_by_outside_corner: Dict[CornerType, TileConvKernel] = {
            CornerType.SW: np.array([[0, 1, 1], [0, 0, 1], [0, 0, 0]]),
            CornerType.SE: np.array([[1, 1, 0], [1, 0, 0], [0, 0, 0]]),
            CornerType.NW: np.array([[0, 0, 0], [0, 0, 1], [0, 1, 1]]),
            CornerType.NE: np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0]]),
        }
        return kernels_by_outside_corner[self]


class FootprintType(Enum):
    RECTANGLE = "RECTANGLE"
    L_SHAPE = "L_SHAPE"
    T_SHAPE = "T_SHAPE"
    IRREGULAR = "IRREGULAR"


class Orientation(Enum):
    HORIZONTAL = "HORIZONTAL"
    VERTICAL = "VERTICAL"

    def other(self) -> "Orientation":
        return Orientation.HORIZONTAL if self == Orientation.VERTICAL else Orientation.VERTICAL


class Azimuth(Enum):
    NORTH = "NORTH"
    EAST = "EAST"
    SOUTH = "SOUTH"
    WEST = "WEST"

    @property
    def convolution_kernel(self) -> TileConvKernel:
        kernels_by_azimuth = {
            Azimuth.NORTH: np.array([[0, 1, 0], [0, 0, 0], [0, 0, 0]]),
            Azimuth.EAST: np.array([[0, 0, 0], [0, 0, 1], [0, 0, 0]]),
            Azimuth.SOUTH: np.array([[0, 0, 0], [0, 0, 0], [0, 1, 0]]),
            Azimuth.WEST: np.array([[0, 0, 0], [1, 0, 0], [0, 0, 0]]),
        }
        return kernels_by_azimuth[self]

    @property
    def aligned_axis(self) -> Axis:
        if self in {Azimuth.NORTH, Azimuth.SOUTH}:
            return Axis.X
        else:
            return Axis.Z

    @property
    def angle_from_positive_x(self) -> int:
        if self == Azimuth.NORTH:
            return -90
        elif self == Azimuth.EAST:
            return 0
        elif self == Azimuth.SOUTH:
            return 90
        elif self == Azimuth.WEST:
            return 180
        else:
            raise SwitchError(self)

    @property
    def opposite(self) -> "Azimuth":
        if self == Azimuth.NORTH:
            return Azimuth.SOUTH
        elif self == Azimuth.EAST:
            return Azimuth.WEST
        elif self == Azimuth.SOUTH:
            return Azimuth.NORTH
        elif self == Azimuth.WEST:
            return Azimuth.EAST
        else:
            raise SwitchError(self)


class WallType(Enum):
    NORTH = "NORTH"
    EAST = "EAST"
    SOUTH = "SOUTH"
    WEST = "WEST"

    @property
    def azimuth(self) -> Azimuth:
        return Azimuth(self.value)

    @property
    def is_vertical(self) -> bool:
        return self in {WallType.EAST, WallType.WEST}

    @property
    def convolution_kernel(self) -> TileConvKernel:
        kernels_by_wall = {
            WallType.NORTH: np.array([[0, 0, 0], [0, 0, 0], [1, 1, 1]]),
            WallType.EAST: np.array([[1, 0, 0], [1, 0, 0], [1, 0, 0]]),
            WallType.SOUTH: np.array([[1, 1, 1], [0, 0, 0], [0, 0, 0]]),
            WallType.WEST: np.array([[0, 0, 1], [0, 0, 1], [0, 0, 1]]),
        }
        return kernels_by_wall[self]

    @property
    def corner_types(self) -> Tuple[CornerType, CornerType]:
        corner_types_by_wall_type = {
            WallType.NORTH: (CornerType.NW, CornerType.NE),
            WallType.EAST: (CornerType.NE, CornerType.SE),
            WallType.SOUTH: (CornerType.SE, CornerType.SW),
            WallType.WEST: (CornerType.NW, CornerType.SW),
        }
        return corner_types_by_wall_type[self]



from typing import ClassVar

import attr

from avalon.datagen.world_creation.entities.tools.tool import Tool


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Weapon(Tool):
    WEAPON_VALUE: ClassVar = 0


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Stick(Weapon):
    WEAPON_VALUE: ClassVar = 1.0
    mass: float = 5.0
    resource_file: str = "res://items/stick.tscn"

    def get_offset(self) -> float:
        return 0.35


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class LargeStick(Stick):
    WEAPON_VALUE: ClassVar = 1.0
    resource_file: str = "res://items/large_stick.tscn"

    def get_offset(self) -> float:
        return 0.35


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Rock(Weapon):
    WEAPON_VALUE: ClassVar = 0.5

    mass: float = 1.0
    resource_file: str = "res://items/rock.tscn"

    def get_offset(self) -> float:
        return 0.25


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class LargeRock(Rock):
    WEAPON_VALUE: ClassVar = 0.5
    resource_file: str = "res://items/large_rock.tscn"

    def get_offset(self) -> float:
        return 0.5
